
"""
This module starts up an HTTP server on the port specified as the first
command-line parameter serving from the subversion path specified as the
second command-line parameter. All files in the specified folder and all
folders underneath it will be served back. The svn:mime-type property is
correctly interpreted and used as the mime type sent back; if it's not present,
a mime-type is chosen based on the file's extension. If a folder is visited,
its index.html file, if any, is sent back. Otherwise, an error is sent.
Listing of files and folders is not yet supported.

If svnweb should authenticate with the subversion server, the username and
password should be specified as the third and fourth command-line parameters,
respectively.

This module was designed to have very high performance. timing wget on a page
2KB in size served from svnweb reports 144 milliseconds taken by wget to
download and save the page. Note, though, that this was with svnweb and the
backing subversion repository on separate machines but the same local network;
serving from a remote subversion repository may be quite a bit slower.

To those of you that might ask why I'm writing this when mod_dav_svn already
provides read access to a repository, the #1 reason is that this script is
considerably easier to run than a full-blown apache install with mod_dav_svn
running inside it. Especially since I run my website, www.opengroove.org,
using Cherokee, which supports reverse-proxying (if you go to
www.opengroove.org, you'll get redirected to www.opengroove.org/static/,
which Cherokee proxies over to an instance of svnweb) but not mod_dav_svn.
"""

import sys
from BaseHTTPServer import BaseHTTPRequestHandler, HTTPServer
from pysvn import Client
from traceback import print_exc
import mimetypes
from mwlib.uparser import simpleparse
from mwlib.xhtmlwriter import MWXHTMLWriter
from xml.etree import ElementTree

mimetypes.init()

error_response = """
<html><body>
There doesn't appear to be a page with that name. Sorry about that.
</html><body>
"""

index_names = ["index.html", "index.xhtml", "index.w"]

def make_index_paths(path):
    return [(path if path[-1:] == "/" else path + "/") 
            + index for index in index_names]

class HTTPHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        path = self.path
        if ".." in path:
            raise Exception("Path can't contain two dots in a row")
        print "Initial path: " + path
        if path == "":
            path = "/"
        if path[0] != "/":
            raise Exception("Path doesn't start with a forward slash. It's "
                    + path)
        path = root_path + path
        check_paths = [path] + make_index_paths(path)
        print "Trying possible paths"
        result = None
        for test_path in check_paths:
            print "Getting result for path " + test_path
            try:
                result = client.cat(test_path)
                path = test_path
                break
            except:
                print "No such file."
        if result == None:
            self.send_response(404)
            self.send_header("Content-Type", "text/html")
            self.no_cache()
            self.end_headers()
            self.wfile.write(error_response)
            return
        print "Got it! Propgetting..."
        # We have the file. Now we go figure out if we're supposed to use an
        # alternate displayer.
        try:
            display_type = client.propget("svnweb:display", path).values()[0]
        except:
            display_type = None
        print "svnweb:display is " + str(display_type)
        if display_type == "mediawiki":
            print "Forwarding to mediawiki renderer"
            mime_type, result = display_mediawiki(self, path, result)
        else:
            print "No known renderer set with svnweb:display, displaying directly"
            try:
                mime_type = client.propget("svn:mime-type", path).values()[0]
            except:
                print "Couldn't propget, looking up dynamically"
                mime_type, _ = mimetypes.guess_type(path)
                if mime_type is None:
                    print "Dynamic lookup didn't find anything. Using text/plain."
                    mime_type = "text/plain"
        print "Mime type is " + mime_type
        self.send_response(200)
        print "Response sent"
        if mime_type != None:
            print "Sending mime type"
            self.send_header("Content-Type", mime_type)
        print "Sending no-cache headers"
        self.no_cache()
        print "Ending headers"
        self.end_headers()
        print "Writing..."
        self.wfile.write(result)
        print "Output sent!"
        # That's it!
    
    def no_cache(self):
        self.send_header("Pragma", "no-cache")
        self.send_header("Cache-Control", "no-cache")
        self.send_header("Expires", "Mon, 22 Nov 2010 01:00:00 GMT")
    
    def address_string(self):
        # This stops the server from looking up the client's hostname,
        # which was hugely slowing things down when I tested it out. We'll
        # just return their IP address instead.
        return str(self.client_address[0])

def get_login(*args):
    return True, username, password, False

def main():
    global port
    global root_path
    global username
    global password
    global client
    if len(sys.argv) <= 2:
        print "You need to specify the port to serve on and the URL under"
        print "which files will be looked up, respectively."
        sys.exit()
    client = Client()
    client.callback_get_login = get_login
    port = int(sys.argv[1])
    root_path = sys.argv[2]
    if root_path[-1] == "/":
        root_path = root_path[:-1]
    if len(sys.argv) <= 4:
        username = None
        password = None
    else:
        username = sys.argv[3]
        password = sys.argv[4]
    print "Port: " + str(port)
    print "Root path: " + root_path
    print "Username: " + str(username)
    print "Password: " + str(("*" * len(password)) if 
            isinstance(password, basestring) else password)
    print
    print "svnweb has successfully started up."
    HTTPServer(("", port), HTTPHandler).serve_forever()


def display_mediawiki(request, path, text):
    article = simpleparse(text)
    try:
        # Try to figure out a title by looking up svn properties
        caption = client.propget("svnweb:mediawiki:title", path).values()[0]
    except:
        caption = None
    if caption is None: # No explicit name, so we'll derive it from the filename
        caption = path.rpartition("/")[2]
        if "." in caption:
            caption = caption.rpartition(".")[0]
    article.caption = caption
    writer = MWXHTMLWriter()
    writer.xwriteStyle = mediawiki_xwriteStyle
    writer.xwriteArticleLink = mediawiki_xwriteArticleLink
    element = writer.write(article)
    result = ElementTree.tostring(element)
    result = """
    <html><head><title>%s</title>
    
    <style type="text/css">
    
    body {font-size: 13px; font-family: sans-serif; padding: 30px;
          background-color: #f9f9f9}
    
    .content {border: 1px solid #aaa; padding: 10px; padding-top: 0px;
              background-color: white}
    
    .content div[class~="mwx.paragraph"] {margin-bottom: 12px}
    
    .content > div > h1 {font-size: 24px; width: 100%%;
                         border-bottom: 1px solid #aaa; margin-bottom: 12px;
                         margin-top: 8px}
    
    .content > div > div > h2 {font-size: 19px; width: 100%%; 
                           border-bottom: 1px solid #aaa;
                           margin-bottom: 8px}
    .content > div > div > div > h2 {font-size: 17px; margin-bottom: 7px;
                                     margin-top: 18px}
    .content > div > div > div > div > h2 {font-size: 15px; margin-bottom: 4px}
    .content > div > div > div > div > div > h2 {font-size: 13px; margin-bottom: 4px}
    
    span[class~="mwx.svnweb.bold"] {font-weight: bold}
    span[class~="mwx.svnweb.italic"] {font-style: italic}
    
    </style>
    
    </head><body>
    <div class="content">
    """ % caption + result + """
    </div></body></html>
    """
    return "text/html", result


def mediawiki_xwriteStyle(style):
    """
    A custom writer function that writes style elements. This is the part that
    generates the html from ''' and '' in mediawiki pages.
    """
    e = ElementTree.Element("span")
    caption_map = {"'''": "mwx.svnweb.bold", "''": "mwx.svnweb.italic"}
    try:
        e.set("class", caption_map[style.caption])
    except KeyError:
        raise Exception("No style information for " + str(style.caption))
    return e

def mediawiki_xwriteArticleLink(obj):
    a = ElementTree.Element("a", href=obj.url or obj.target or "#")
    a.set("class", "mwx.link.article")
    if not obj.children:
        a.text = obj.target
    return a




















"""
Resolvers have the signature resolver(source, type, name) -> answer

type is the string type of the request ("A", "MX", etc)

name is the request itself ("www.example.com" etc)

answer is a list of Answer instances
"""

import socket as s
import dns.message
import dns.rrset
from dns.rdtypes.IN import *
from dns.rdtypes.ANY import *
from dns.rdataclass import from_text as text_to_class
from dns.rdatatype import from_text as text_to_type
from afn.dnsd import data as d

class Server(object):
    def __init__(self, host="0.0.0.0", port=53, socket=None):
        if socket:
            self.socket = socket
        else:
            self.socket = s.socket(s.AF_INET, s.SOCK_DGRAM)
            self.socket.bind((host, port))
        self.resolvers = []
        self.filters = []
    
    def run(self):
        while True:
            message_content, (source_host, source_port) = self.socket.recvfrom(8192)
            print "DNSD: Message is " + repr(message_content)
            message = dns.message.from_wire(message_content)
            print "DNSD: Request: " + repr(message)
            response = self.process_message((source_host, source_port), message)
            response_content = response.to_wire()
            print "DNSD: Response: " + repr(response)
            print "DNSD: Response is " + repr(response_content)
            self.socket.sendto(response_content, (source_host, source_port))
    
    def process_message(self, source, message):
        questions = message.question
        answers = [self.process_question(source, q) for q in questions]
        answers = [answer for answer_list in answers for answer in answer_list]
        response = dns.message.make_response(message)
        response.answer += answers
        return response
    
    def process_question(self, source, question):
        if dns.rdataclass.to_text(question.rdclass) != "IN":
            raise Exception("Only IN class queries are supported.")
        datatype = dns.rdatatype.to_text(question.rdtype)
        name = question.name
        textual_name = ".".join(question.name.labels[:-1])
        answers = None
        for resolver in self.resolvers:
            answers = resolver(source, datatype, textual_name)
            if answers:
                break
        if not answers:
            raise Exception("No sources could answer " + repr(question))
        return data_to_dnspy(answers)


def data_to_dnspy(data):
    if isinstance(data, list):
        return [data_to_dnspy(x) for x in data]
    if isinstance(data, d.Answer):
        result = dns.rrset.RRset(dns.name.from_text(data.name),
                text_to_class("IN"), text_to_type(data.type))
        result.ttl = data.ttl
        result.items += data_to_dnspy(data.items)
        return result
    if isinstance(data, d.A):
        return A.A(text_to_class("IN"), text_to_type("A"), data.ip)
    if isinstance(data, d.AAAA):
        return AAAA.AAAA(text_to_class("IN"), text_to_type("AAAA"), data.ip)
    if isinstance(data, d.CNAME):
        return CNAME.CNAME(text_to_class("IN"), text_to_type("CNAME"),
                dns.name.from_text(data.target))
    raise Exception("Can't convert instance of " + repr(type(data)) + ": " + repr(data))
                






























            




from autobus2 import Bus
from autobus2.filter import NotEqualTo
from afn.utils.concurrent import dump_thread_traces

class ChatService(object):
    def message_happened(self, sender, message):
        print sender + ": " + message

def signed_on(proxy, connection, info):
    print info["name"] + " has signed on."

def signed_off(proxy, connection, info):
    print info["name"] + " has signed off."

def main():
    name = raw_input("Enter your name: ")
    print "You've been successfully connected to the chat network."
    print "Type messages to send them."
    with Bus() as bus:
        # Publish ChatService to allow messages to be sent to us
        service = bus.create_service({"autobus.example": "chat1", "name": name},
                from_py_object=ChatService())
        # Connect to other chat services on the network so that we can send
        # messages to them
        with bus.get_service_proxy({"autobus.example": "chat1",
                "service": NotEqualTo(service.id)}, bind_function=signed_on,
                unbind_function=signed_off, multiple=True) as proxy:
            # Loop until the user hits Ctrl+C
            while True:
                try:
                    message = raw_input()
                except KeyboardInterrupt:
                    return
                if message == "==trace==":
                    dump_thread_traces()
                    continue
                proxy["message_happened"](name, message, callback=None)








import sys
from libautobus import AutobusConnection
from optparse import OptionParser
from time import sleep

def main():
    try:
        import pynotify
    except ImportError:
        print "You don't have the Python libnotify binding installed."
        sys.exit()
    
    if not pynotify.init("timernotify"):
        print "Libnotify could not be initialized."
        sys.exit()
    
    parser = OptionParser(usage="usage: python -m timernotify [options]", 
            add_help_option=False)
    parser.add_option("-h", type="string", action="store", dest="host", default="localhost",
            help="The host that the Autobus server is running on. Without this "
            "option, localhost will be used.")
    parser.add_option("-s", action="store_const", const=True, dest="state_change",
            default=False, help="If present, all state changes (instead of just "
            "when the timer beeps) will cause a notification to be shown.")
    parser.add_option("-?", "--help", action="help")
    
    options, command_line_args = parser.parse_args()
    
    bus = AutobusConnection(host=options.host)
    def beeping_listener(timer):
        pynotify.Notification("Timer " + str(timer), "is beeping.").show()
    
    def state_change_listener(timer, state):
        state_string = {1: "counting up", 2: "counting down", 3: "stopped"}[state]
        pynotify.Notification("Timer " + str(timer), "is now " + state_string + 
                ".").show()
    
    bus.add_event_listener("timer", "beeping", beeping_listener)
    if options.state_change:
        bus.add_event_listener("timer", "manual_state_change", state_change_listener)
    
    bus.connect()
    
    try:
        while True:
            sleep(1)
    except KeyboardInterrupt:
        print "Interrupted, shutting down"
    finally:
        bus.shutdown()


















from datetime import datetime, timedelta as interval
from time import sleep
from autobus2 import Bus
import sys
from concurrent import as_new_thread
from afn.utils import speakutils, timeutils

class RPC(object):
    """
    Provides functions for saying the current time. This program says the time
    automatically every 15 minutes as well. There isn't yet a way to configure
    this automatic time-saying from Autobus, but I'll be adding that soon. 
    """
    @as_new_thread
    def time(self):
        """
        Instructs the speak server to say the current time.
        """
        speak_server["say_text"]("the_time is " + 
                speakutils.format_time(current_time), callback=None)
    
    @as_new_thread
    def datetime(self):
        """
        Instructs the speak server to say the current date and time.
        """
        speak_server["say_text"]("today is " + 
                speakutils.format_weekday_month_day(current_time) + 
                " :p:600 the_time is " + 
                speakutils.format_time(current_time), callback=None)
    
    def current_time(self):
        """
        Returns the current minute according to the server. This is the time
        that would be spoken had datetime() or time() been called instead of
        this function.
        """
        return timeutils.to_time(current_time)

def main():
    global current_time
    global speak_server
    global rpc
    rpc = RPC()
    
    with Bus() as bus:
        service = bus.create_service({"type": "saytime"}, from_py_object=rpc)
        speak_server = bus.get_service_proxy({"type": "speak"}, multiple=True)
        
        start_time = datetime.now()
        current_time = start_time
        next_minute = start_time
    
        while True:
            interval_to_next = next_minute - datetime.now()
            time_to_next = float(interval_to_next.seconds) + (interval_to_next.microseconds / 1000000.0)
            if interval_to_next.days < 0:
                time_to_next = 0
            print "Waiting " + str(time_to_next) + " seconds"
            sleep(time_to_next)
            sleep(0.2) # Make sure we're at the next minute
            current_hour = next_minute.hour
            current_minute = next_minute.minute
            current_time = next_minute
            next_minute = next_minute + interval(seconds=70)
            next_minute = datetime(year=next_minute.year, month=next_minute.month, day=next_minute.day, hour=next_minute.hour, minute=next_minute.minute)
            print "Processing instructions for " + str(current_hour) + ":" + str(current_minute)
            if (current_hour >= 8 and current_hour < 20) or (current_hour == 20 and 
                                                             current_minute == 0): 
                if current_minute == 0:
                    rpc.datetime()
                elif (current_minute % 15) == 0:
                    rpc.time()


















from librtk.protocols import ThreadedProtocol
from librtkclient import Connection
import librtkinter
from librtk.constants import DEFAULT_PORT
from socket import socket as Socket, error as SocketError
import Tkinter as tkinter
import sys
from urlparse import urlparse
from Queue import Queue, Empty
from traceback import print_exc
from utils import print_exceptions
from functools import partial

def main():
    if len(sys.argv) <= 1:
        print "You need to specify the url of the application to display."
        return
    scheme, location = urlparse(sys.argv[1])[0:2]
    if scheme != "rtk":
        print ("Only rtk:// urls are supported for now. I might add an HTTP "
                "transport at some point in the future.")
    if ":" not in location:
        location += ":" + str(DEFAULT_PORT)
    host, port = location.split(":")
    port = int(port)
    socket = Socket()
    try:
        socket.connect((host, port))
    except SocketError, e:
        print "Error while connecting: " + str(e)
        return
    protocol = ThreadedProtocol(socket)
    connection, tk = librtkinter.start_connection(protocol)
    print "Started up!"
    try:
        tk.mainloop()
    except KeyboardInterrupt:
        print "Interrupted, shutting down"
        connection.close()
    print "Terminated."
    



























__doc__ = """\
FIXME: THIS HAS BEEN PORTED TO AUTOBUS 2 BUT NOT YET TESTED. IT NEEDS TO BE
TESTED AND ANY BUGS WORKED OUT BEFORE IT CAN BE USED AGAIN.

Autoconfigure is a library that Python applications connecting to Autobus and
providing interfaces can use. It allows remote editing of configuration files
that a particular application uses. Typically, an application called "example"
will instruct Autoconfigure to register itself as configure.example for the
application's normal configuration file.

If the application is written to support this, changes made by remote
applications to configuration can be applied as soon as they are made. It's
really as simple as having the application register appropriate listeners on
individual configuration variables or the entire configuration system.
"""

sections_object_doc = """\

"""

options_object_doc = """\
"""

"""
Ok so, how this is going to work. Basically, we load up the configuration file
and represent it as a RawConfigParser. We load up the changes from the
specified configuration file, or create it if it doesn't exist. We then
register ourselves as a service on the specified Autobus bus under
the specified name. All of the functions lock on a variable, including the
ones made available to the program that's using us. When a function's called
that changes data, the data's changed in the config parser, the configuration
file's written out, and all relevant listeners are called.

Functions that we need to provide:

sections()
add_section(name)
has_section(name)
options(section)
has_option(section, option)
get(section, option)
getint(section, option)
getfloat(section, option)
getboolean(section, option)
items(section)
set(section, option, value)
remove_option(section, option)
remove_section(section)
"""

from ConfigParser import RawConfigParser
from threading import RLock
from concurrent import synchronized

class _AutoconfigureService(object):
    """
    This is an interface provided by the Autoconfigure system. Autoconfigure
    is a Python library (although it could feasibly be converted into a
    standalone program usable from other programming languages) that lets
    remote Autobus clients edit configuration files. Each Autoconfigure
    interface registered on the Autobus server edits one configuration file.
    Convention is that a program registering the interface example will
    register its configuration file under the interface configure.example.
    
    If the program embedding Autoconfigure supports it (and most do), changes
    made via this interface will take effect immediately in the program
    without requiring a restart.
    
    The documentation for the functions on this interface is pulled directly
    from Python's RawConfigParser. If there are any Python-specific references
    in there, that's why.
    """
    def __init__(self, config):
        self.config = config
    
    def sections(self):
        return self.config.sections()
    
    def add_section(self, name):
        return self.config.add_section(name)
    
    def has_section(self, name):
        return self.config.has_section(name)
    
    def options(self, section):
        return self.config.options(section)
    
    def has_option(self, section, option):
        return self.config.has_option(section, option)
    
    def get(self, section, option):
        return self.config.get(section, option)
    
    def getint(self, section, option):
        return self.config.getint(section, option)
    
    def getfloat(self, section, option):
        return self.config.getfloat(section, option)
    
    def getboolean(self, section, option):
        return self.config.getboolean(section, option)
    
    def items(self, section):
        return self.config.items(section)
    
    def set(self, section, option, value):
        return self.config.set(section, option, value)
    
    def remove_option(self, section, option):
        return self.config.remove_option(section, option)
    
    def remove_section(self, section):
        return self.config.remove_section(section)
    
    # Now we'll copy the docs from RawConfigParser
    for name, function in locals().items()[:]:
        if not name.startswith("_"):
            function.__doc__ = getattr(RawConfigParser, name).__doc__
    del name
    del function

# TODO: convert this to one lock per Configuration instance at
# some point in the future
_lock = RLock()

class Configuration(object):
    def __init__(self, bus, target_name, file_name):
        self.bus = bus
        self.target_name = target_name
        self.file_name = file_name
        self.config = RawConfigParser()
        self.config.read(file_name)
        if target_name is not None:
            self.service = bus.create_service({"type": "autoconfigure", "target": target_name},
                    active=False, use_py_object=_AutoconfigureService(self))
            self.sections_object = self.service.create_object("sections",
                    self._compute_sections(), doc=sections_object_doc)
            self.options_object = self.service.create_object("options",
                    self._compute_options(), doc=options_object_doc)
    
    @synchronized(_lock)
    def _save(self):
        with open(self.file_name, "w") as file:
            self.config.write(file)
        self.sections_object.set_value(self._compute_sections())
        self.options_object.set_value(self._compute_options())
    
    @synchronized(_lock)
    def _compute_sections(self):
        return self.config.sections()
    
    @synchronized(_lock)
    def _compute_options(self):
        return dict((section, dict(self.config.items(section)))
                for section in self.config.sections())

    @synchronized(_lock)
    def sections(self):
        return self.config.sections()
    
    @synchronized(_lock)
    def add_section(self, name):
        self.config.add_section(name)
        self._save()
    
    @synchronized(_lock)
    def has_section(self, name):
        return self.config.has_section(name)
    
    @synchronized(_lock)
    def options(self, section):
        return self.config.options(section)
    
    @synchronized(_lock)
    def has_option(self, section, option):
        return self.config.has_option(section, option)
    
    @synchronized(_lock)
    def get(self, section, option):
        return self.config.get(section, option)
    
    @synchronized(_lock)
    def getint(self, section, option):
        return self.config.getint(section, option)
    
    @synchronized(_lock)
    def getfloat(self, section, option):
        return self.config.getfloat(section, option)
    
    @synchronized(_lock)
    def getboolean(self, section, option):
        return self.config.getboolean(section, option)
    
    @synchronized(_lock)
    def items(self, section):
        return self.config.items(section)
    
    @synchronized(_lock)
    def set(self, section, option, value):
        self.config.set(section, option, value)
        self._save()
    
    @synchronized(_lock)
    def remove_option(self, section, option):
        self.config.remove_option(section, option)
        self._save()
    
    @synchronized(_lock)
    def remove_section(self, section):
        self.config.remove_section(section)
        self._save()
        






























from afn.utils import timeutils

def format_time(time):
    """
    Formats a date into a string representing the time of the date that
    can be included in a sentence. For example,
    time_format_time(datetime(hour=15, minute=14)) would
    return ":t:3:14pm".
    """
    hour, minute = time.hour, time.minute
    period = "am" if hour < 12 or hour == 24 else "pm"
    hour = hour % 12
    if hour == 0:
        hour = 12
    return ":t:" + str(hour) + ":" + str(minute).rjust(2, "0") + period

def format_month_day(time):
    """
    Formats a date into a string representing the month and day of the date
    that can be included in a sentence. For example,
    time_format_month_day(datetime(month=3, day=14)) would
    return "march :n:14".
    """
    month = ["january", "february", "march", "april",
             "may", "june", "july", "august",
             "september", "october", "november", "december"
             ][time.month - 1]
    day = str(time.day)
    return month + " :n:" + day

def format_weekday(time):
    """
    Formats a date into a string representing the day of the week that the
    day is on.
    """
    return time.strftime("%A").lower()

def format_weekday_month_day(time):
    """
    Returns time_format_weekday(time) + " " + time_format_month_day(time)
    """
    return (format_weekday(time) + " " + 
            format_month_day(time))



"""
Autobus is a decentralized RPC system and service discovery system. You might
think of it as a mix of Bonjour and D-Bus, with a bit of JSON-RPC thrown in for
good measure.

With Autobus, various programs on a single computer or spread out over a
network can create, publish, and find each other's services. Services are
simply collections of functions that can be called, events that can be listened
for, and objects whose values can be watched. (Other types of things in
addition to those three are planned for the future.)

You're probably still a bit confused by what Autobus exactly does, so let's get
into some examples. Let's start off with the venerable Hello, World example:

# server.py
from autobus2 import Bus, wait_for_interrupt
class HelloService(object):
    def hi(self, text="world"):
        print "Saying hi to " + str(text)
        return "Hi, " + str(text) + "! How are you?"
with Bus() as bus:
    bus.create_service({"autobus.example": "hello_server"}, from_py_object=HelloService())
    wait_for_interrupt()

# client.py
from autobus2 import Bus
with Bus() as bus:
    with bus.get_service_proxy({"autobus.example": "hello_server"}) as service:
        service.wait_for_bind(timeout=2)
        print service["hi"]("great big round world")

Run server.py first, then run client.py. You'll see the server print out
"Saying hi to world" and the client print "Hello, world! How are you?". Try
running client.py first; it'll sit there waiting until you start up server.py,
at which point it will happily print out its greetings. Now try running
server.py on a completely different machine on the same local network; both
machines (one running server.py and the other running client.py) happily print
out their salutations, even though you didn't instruct, for example, client.py
which host on your network was running server.py.

So how did the client find out about the server? That's the beauty of Autobus;
it comes with service discovery built in. Notice in the client, when we call
bus.get_service_proxy (which, by the way, asks Autobus to get us an object that
knows how to find other services on the network; we'll call these service-finding
things proxies from now on), we pass it a dictionary. Notice also that when we
call bus.create_service in the server, we pass it the same dictionary.

That's how the client's finding the server; it searches the network for another
Autobus service whose dictionary (we call this the info dictionary, or the
info object, by the way) contains the keys specified in the call to get_proxy
with the values specified. (Any service containing /at least/ those keys will
match, so, for example, passing {} to bus.get_proxy would cause it to choose
any old Autobus service on the network at random)

TODO: finish this up, and go into more detail about other things like events
and objects and multiple service proxies and individual connections and
discoverers and publishers and such.
"""

import sys
from traceback import print_exc
from autobus2 import net, discovery, local, remote, exceptions, messaging, common, proxy
from autobus2.filter import filter_matches, ANY, NOT_PRESENT
from threading import Thread, RLock
from socket import socket as Socket, error as SocketError, timeout as SocketTimeout, gethostname
from Queue import Queue
from concurrent import synchronized
import time
from utils import no_exceptions, print_exceptions
from afn.utils import singleton
import __builtin__
from concurrent import synchronized_on
from afn.utils import field
try:
    from collections import OrderedDict as _OrderedDict
except ImportError:
    from afn.backports.ordered_dict import OrderedDict as _OrderedDict

SYNC = singleton.Singleton("autobus2.SYNC")
THREAD = singleton.Singleton("autobus2.THREAD")
ASYNC = singleton.Singleton("autobus2.ASYNC")
DISCOVERED = singleton.Singleton("autobus2.DISCOVERED")
UNDISCOVERED = singleton.Singleton("autobus2.UNDISCOVERED")
CHANGED = singleton.Singleton("autobus2.CHANGED")


class DiscoveredService(object):
    @field
    def locations(self):
        """
        An ordered dictionary containing the locations at which this service
        is available. The keys are (host, port) tuples representing the
        locations and the values are lists of the Discoverer instances that
        discovered those particular locations.
        """
    
    @field
    def info(self):
        """
        The info object for this service.
        """
    
    def __init__(self, info):
        self.locations = _OrderedDict()
        self.info = info


class Bus(common.AutoClose):
    def __init__(self, default_discoverers=True, default_publishers=True,
                 port=None):
        """
        Creates a new bus. The bus will listen on the specified port; if none
        is specified (which is the usual case), a port will be chosen from the
        ports not currently in use on this computer.
        """
        self.context_enters = 0
        if port is None:
            port = 0
        self._introspector = None
        self.closed = False
        self.server = Socket()
        self.server.bind(("", port))
        self.server.listen(100)
        self.port = self.server.getsockname()[1]
        self.lock = RLock()
        self.local_services = {}
        self.discovered_services = {}
        self.discovery_listeners = []
        self.service_listeners = []
        self.bound_connections = set()
        self.discoverers = set()
        self.publishers = set()
        if default_discoverers:
            self.install_discoverer(discovery.BroadcastDiscoverer())
        if default_publishers:
            self.install_publisher(discovery.BroadcastPublisher())
        Thread(name="autobus2.Bus.accept_loop", target=self.accept_loop).start()
        # Disable the introspection service for now. I'm seeing what would
        # happen if I have per-service introspection functions and objects, so
        # I'm disabling the bus-wide introspection service.
        # self._create_introspection_service()
    
    def accept_loop(self):
        self.server.settimeout(1)
        while not self.closed:
            try:
                socket = None
                socket = self.server.accept()[0]
                self.setup_inbound_socket(socket)
            except SocketTimeout: # This happens when we time out, which is
                # normal. The 4-second timeout is to fix what appears to be a
                # bug with Windows not properly throwing an exception from
                # accept when another thread closes the socket.
                pass
            except: # This happens when the server socket is closed
                if socket:
                    socket.close() # Make sure it's /really/ closed on the
                    # off chance that something else caused the exception
                if not issubclass(sys.exc_type, SocketError): # Something else
                    # happened
                    print_exc()
                # print "Bus server died"
                return
    
    @synchronized_on("lock")
    def create_service(self, info, active=None, from_py_object=None, doc=""):
        """
        Creates a new service on this bus. info is the info object to use for
        this service. active is True to publish this service immediately, False
        to wait until the returned service's activate() method is called. If
        active isn't specified, it defaults to False if from_object is None and
        True if from_object is not None.
        
        The return value is an instance of local.LocalService. It has methods
        such as create_function that allow functions, events, objects, and such
        to be created on the service.
        """
        if active is None:
            if from_py_object is not None:
                active = True
            else:
                active = False
        # Create a new id for the service
        service_id = messaging.create_service_id()
        self.set_remote_info_builtins(service_id, info)
        # Create the actual service object
        service = local.LocalService(self, service_id, info, doc)
        # Then store the service in our services map
        self.local_services[service_id] = service
        if from_py_object is not None:
            service.use_py_object(from_py_object)
        service._add_introspection()
        # If the service is to be immediately activated it, then we should
        # do so
        if active:
            service.activate()
        return service
    
    def _close_service(self, service):
        # The service will already be deactivated when this gets called, so we
        # don't need to remove it from any of the publishers. So I think the
        # only thing we need to do is remove it from our map of services.
        del self.local_services[service.id]
        self._i_update(service.id)
    
    @synchronized_on("lock")
    def setup_inbound_socket(self, socket):
        # Create a connection and then add it to our list of connections
        connection = local.RemoteConnection(self, socket)
        self.bound_connections.add(connection)
    
    def connect(self, host, port, service_id, timeout=10, open_listener=None,
                close_listener=None, fail_listener=None, lock=None):
        """
        Opens a connection to the specified service on the specified host/port.
        
        The connection will be returned immediately. The actual connection to
        the server will be made as soon as possible in the future. If you need
        to block until the connection actually connects, call wait_for_connect
        on the returned Connection object.
        
        The connection will attempt to reconnect indefinitely whenever it is
        disconnected. If you don't want this behavior, specify a close_listener
        that calls the connection's close method.
        
        Timeout is the TCP timeout to use when connecting. The default is 10;
        this is usually a suitable default. You'll probably only want to
        increase this if you're working on a particularly latent network.
        
        open_listener and close_listener are functions accepting one argument.
        They will be called when the connection successfully connects and when
        the connection disconnects, respectively, and the connection itself
        will be passed in. They are both run synchronously on the connection's
        input thread, so it's guaranteed that, for example, the connection will
        not attempt to reconnect until close_listener has returned. Thus
        close_listener could be set to a function that just closes the
        specified connection in order to effectively disable the auto-reconnect
        feature of connections.
        """
        return remote.Connection(self, host, port, service_id, timeout, open_listener, close_listener, fail_listener, lock)
    
    def connect_to(self, info_filter, timeout=10, open_listener=None, close_listener=None, fail_listener=None, lock=None):
        with self.lock:
            for service_id, d in self.discovered_services.items():
                if filter_matches(d.info, info_filter):
                    host, port = d.locations.keys()[0]
                    return self.connect(host, port, service_id, timeout, open_listener, close_listener, fail_listener, lock)
            raise exceptions.NoMatchingServiceException()
    
    def get_service_proxy(self, info_filter, bind_function=None, unbind_function=None, multiple=False):
        with self.lock:
            if multiple:
                return proxy.MultipleServiceProxy(self, info_filter, bind_function, unbind_function)
            else:
                return proxy.SingleServiceProxy(self, info_filter)
    
    @synchronized_on("lock")
    def close(self):
        self.closed = True
        # First we shut down all of our discoverers
        for discoverer in self.discoverers:
            discoverer.shutdown()
        # Then we need to unpublish all of our services and shut down all
        # of our publishers
        for publisher in self.publishers:
            for service in self.local_services.values():
                if service.active:
                    publisher.remove(service)
            publisher.shutdown()
        # Then we shut down the server socket
        net.shutdown(self.server)
        # Then we close all of the connections currently connected to us
        for c in self.bound_connections:
            with no_exceptions:
                c.close()
        # And that's it!
        # TODO: In the future, store some sort of closed field so that if
        # someone tries to double-close us, we can tell and just ignore it
        # the second time
    
    @synchronized_on("lock")
    def install_publisher(self, publisher):
        # Add the publisher to our list and start it up
        self.publishers.add(publisher)
        publisher.startup(self)
        # Then register all of our local services with the publisher
        for service in self.local_services.values():
            if service.active:
                publisher.add(service)
    
    @synchronized_on("lock")
    def remove_publisher(self, publisher):
        # Check to make sure that the publisher is already installed
        if publisher not in self.publishers:
            raise __builtin__.ValueError("The specified publisher is not currently installed on this bus.")
        # Remove the publisher from our list of publishers
        self.publishers.remove(publisher)
        # Unpublish all of our services from the publisher
        for service in self.local_services.values():
            if service.active:
                publisher.remove(service)
        # Then we shut down the publisher
        publisher.shutdown()
    
    @synchronized_on("lock")
    def install_discoverer(self, discoverer):
        # Add the discoverer to our list of discoverers, then start it up
        self.discoverers.add(discoverer)
        discoverer.startup(self)
    
    @synchronized_on("lock")
    def remove_discoverer(self, discoverer):
        # Check to make sure that the discoverer has already been installed
        if discoverer not in self.discoverers:
            raise __builtin__.ValueError("The specified discoverer is not currently installed on this bus.")
        # Remove the discoverer from our list of discoverers, then shut it
        # down
        self.discoverers.remove(discoverer)
        discoverer.shutdown()
    
    def set_local_info_builtins(self, host, port, service_id, info):
        new_info = info.copy()
        new_info["host"] = host
        new_info["port"] = port
        new_info["service"] = service_id
        return new_info
    
    def set_remote_info_builtins(self, service_id, info):
        info["hostname"] = gethostname()
    
    @synchronized_on("lock")
    def discover(self, discoverer, host, port, service_id, info):
        # print "Discovered:", (host, port, service_id, info)
        info = self.set_local_info_builtins(host, port, service_id, info)
        # Check to see if the specified service has been discovered yet, and if
        # it hasn't, create an entry for it
        is_new_service = False
        if service_id not in self.discovered_services:
            self.discovered_services[service_id] = DiscoveredService(info)
            is_new_service = True
        discovered_service = self.discovered_services[service_id]
        # Check to see if the specified host/port combination is already
        # present, and if it isn't, add it.
        if (host, port) not in discovered_service.locations:
            discovered_service.locations[(host, port)] = []
        discoverer_list = discovered_service.locations[(host, port)]
        # Check to see if this discoverer has already discovered that host/port
        if discoverer in discoverer_list:
            print ("Warning: discoverer " + str(discoverer) + 
                   " tried to rediscover " + str((host, port, service_id)) +
                   " with info " + str(info))
            return
        # It hasn't, so add it.
        discoverer_list.append(discoverer)
        # The check to see if we need to notify listeners, and do so if we
        # need to
        if is_new_service:
            self.notify_service_listeners(service_id, host, port, info, DISCOVERED) 
    
    @synchronized_on("lock")
    def undiscover(self, discoverer, host, port, service_id):
        # print "Undiscovered:", (host, port, service_id)
        # Check to see if the specified service has been discovered.
        if service_id not in self.discovered_services:
            print ("Warning: discoverer " + str(discoverer) + " tried to "
                   "undiscover " + str((host, port, service_id)) + " when "
                   "such a service does not exist.")
            return
        discovered_service = self.discovered_services[service_id]
        if (host, port) not in discovered_service.locations:
            print ("Warning: discoverer " + str(discoverer) + " tried to "
                   "undiscover " + str((host, port, service_id)) + " when "
                   "that host/port has not yet been discovered.")
            return
        discoverer_list = discovered_service.locations[(host, port)]
        if discoverer not in discoverer_list:
            print ("Warning: discoverer " + str(discoverer) + " tried to "
                   "undiscover " + str((host, port, service_id)) + " when "
                   "this discoverer hasn't discovered that host/port yet.")
            return
        discoverer_list.remove(discoverer)
        if not discoverer_list:
            if discovered_service.locations.keys()[0] == (host, port): # We're
                # removing the first (and therefore default) location, so if
                # there's another location, we need to let the service
                # listeners know that there's a new default location
                if len(discovered_service.locations) > 1: # There will be
                    # another location even after we delete this one
                    new_host, new_port = discovered_service.locations.keys()[1]
                    if not self.closed: # Don't issue changes if we're shutting down
                        self.notify_service_listeners(service_id, new_host, new_port, discovered_service.info, CHANGED)
            del discovered_service.locations[(host, port)]
            if not discovered_service.locations: # That was the last location
                # available for this service, so we delete the service itself,
                # and notify listeners that it was deleted
                del self.discovered_services[service_id]
                self.notify_service_listeners(service_id, host, port, discovered_service.info, UNDISCOVERED)
    
    @synchronized_on("lock")
    def add_service_listener(self, listener, info_filter=None, initial=False):
        """
        Listens for changes in services that are available. listener is a
        function listener(service_id, host, port, info, event) which will be
        called whenever a service becomes available, a service disappears, or
        the host/port that should be used to access a particular service
        changes. service_id is the id of the service; host/port is the host/port
        at which the service can be found, info is the service's info object,
        and event is one of DISCOVERED, UNDISCOVERED, or CHANGED.
        
        If info_filter is a dictionary, only services with info objects matching
        that particular filter (as per the filter_matches function) will cause
        the listener to be called. If info_filter is None (the default), or the
        empty dictionary (since all info objects match the empty dictionary),
        the listener will be called for all services.
        
        If initial is True, the listener will be immediately (and synchronously)
        called once for each service that already exists, passing in DISCOVERED
        as the event. Otherwise, the listener will only be called once the next
        
        """
        # Add the listener to our list of listeners
        self.service_listeners.append((info_filter, listener))
        # Check to see if we're supposed to notify the listener about all
        # matching services that already exist
        if initial:
            # Scan all of the services
            for service_id, discovered_service in self.discovered_services.items():
                if filter_matches(discovered_service.info, info_filter):
                    # If this service matches, notify the listener about it
                    host, port = discovered_service.locations.keys()[0]
                    with print_exceptions:
                        listener(service_id, host, port, discovered_service.info, DISCOVERED)
    
    @synchronized_on("lock")
    def remove_service_listener(self, listener, initial=False):
        # Scan the list of listeners and remove this one. Inefficient, it's
        # true, and I hope to make it more efficient later on.
        for index, (info_filter, l) in enumerate(self.service_listeners[:]):
            # See if we've hit the right listener
            if l == listener:
                # If we have, remove the listener
                del self.service_listeners[index]
                if initial:
                    # Scan through the list of services
                    for service_id, discovered_service in self.discovered_services.items():
                        if filter_matches(discovered_service.info, info_filter):
                            # This service matched, so we notify this
                            # listener that the service was removed
                            with print_exceptions:
                                listener(service_id, None, None, None, UNDISCOVERED)
                # We've found our listener and deleted it, so we return now
                return
    
    def notify_service_listeners(self, service_id, host, port, info, event):
        for filter, listener in self.service_listeners:
            if filter_matches(info, filter):
                with print_exceptions:
                    listener(service_id, host, port, info, event)
    
    def _create_introspection_service(self):
        service = self.create_service({"type": "autobus"})
        self._introspector = service
        service.create_object("details", self._i_details_function)
        service.activate()
        self._i_update(None)
    
    def _i_update(self, service_id):
        if self._introspector:
            introspector = self._introspector
            if introspector.objects["details"]:
                introspector.objects["details"].changed()
        if service_id in self.local_services:
            service = self.local_services[service_id]
            if "autobus.details" in service.objects:
                service.objects["autobus.details"].changed()
    
    @synchronized_on("lock")
    def _i_details_function(self):
        services = {}
        for id, service in self.local_services.items():
            details = service._i_details_function()
            services[id] = details
        return services


def wait_for_interrupt():
    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        # print "Interrupted"
        pass
        

































import re
import time

__doc__ = """\
This module contains functions that can be used to extract CPU usage info and
memory usage info from /proc/stat and /proc/meminfo. For /proc/stat
specifically, this module stores the old values obtained on the last read and
uses those to compute the processor's usage since then. It also tracks
multiple processors independently.
"""

class CPUMonitor(object):
    """
    A class that monitors CPU usage via /proc/stat. Each instance of
    CPUMonitor keeps track of the values at the last read and uses these to
    compute the processor's usage since then.
    """
    types = (("user", 0), ("user_nice", 1), ("system", 2))
    
    def __init__(self):
        self.last_map = {} # Maps CPUs to maps of types to their last values
        self._populate(self.last_map) # Initial population
        self.last_time = time.time()
        self.results = {}
        for number in self.last_map:
            self.results[number] = dict((type, 0.0) for type, _ in self.types)
    
    def _populate(self, map):
        with open("/proc/stat") as stat:
            for line in stat:
                if re.match("cpu[0-9] .*", line):
                    number, params = re.match("cpu([0-9]) +(.*)",
                            line).groups()
                    number = int(number)
                    params = params.split(" ")
                    map[number] = dict((key, int(params[index])) for key, index
                            in CPUMonitor.types)
    
    def refresh(self):
        """
        Reads /proc/stat and computes the CPU usage since the last time this
        function was called. The computed usage will then be available through
        all of the get_* functions. The results range from 0.0 to 100.0.
        """
        current_map = {}
        results = {}
        new_time = time.time()
        time_difference = new_time - self.last_time
        self.last_time = new_time
        self._populate(current_map)
        for number, this_map in current_map.items():
            last_map = self.last_map[number]
            results[number] = dict((type, (this_map[type] - last_map[type]) / time_difference)
                    for type, _ in CPUMonitor.types)
        self.last_map = current_map
        self.results = results
    
    def get(self, processor=None):
        """
        Returns a tuple (user, user_nice, system) for the specified processor.
        If processor is None, stats for all of the CPUs will be averaged and
        returned.
        """
        if processor is None:
            results = dict((type, 0.0) for type, _ in CPUMonitor.types)
            for cpu_results in self.results.values():
                for type, _ in CPUMonitor.types:
                    results[type] += cpu_results[type]
            for type, _ in CPUMonitor.types:
                results[type] /= len(self.results)
        else:
            results = self.results[processor]
        return results["user"], results["user_nice"], results["system"]
    
    def get_condensed(self, processor=None):
        """
        Returns a tuple (user, system), where user is user and user_nice
        added together.
        """
        user, user_nice, system = self.get(processor)
        return (user + user_nice), system
    
    def get_processors(self):
        """
        Returns the number of processors present on this system.
        """
        return len(self.results)

class MemoryMonitor(object):
    types = ["MemTotal", "MemFree", "SwapTotal", "SwapFree", "Buffers", "Cached"]
    def __init__(self):
        self.map = dict((key, 0) for key in MemoryMonitor.types)
        self.refresh()
    
    def refresh(self):
        with open("/proc/meminfo") as meminfo:
            for line in meminfo:
                if ":" not in line:
                    continue
                type = line[:line.find(":")]
                if type not in MemoryMonitor.types:
                    continue
                line = line[line.find(":") + 1:].strip()
                linesplit = line.split(" ")
                number = linesplit[0]
                specifier = None
                if len(linesplit) > 1:
                    specifier = linesplit[1].lower()
                number = int(number)
                if specifier == "kb":
                    number *= 1000
                elif specifier == "mb":
                    number *= 1000000
                self.map[type] = number
    
    def get(self):
        """
        returns a tuple (resident, buffers, cached, free, total, swap,
        swapfree, swaptotal) of memory statistics. Each item is an integer (or
        a long) representing the number of bytes.
        """
        total, free, swaptotal, swapfree, buffers, cached = (
                self.map["MemTotal"], self.map["MemFree"],
                self.map["SwapTotal"], self.map["SwapFree"],
                self.map["Buffers"], self.map["Cached"])
        swap = swaptotal - swapfree
        resident = total - (free + buffers + cached)
        return (resident, buffers, cached, free, total, swap,
                swapfree, swaptotal)








from autobus2 import Bus, wait_for_interrupt

def changed(*args):
    print "CHANGED: " + str(args)

def main():
    with Bus() as bus:
        with bus.get_service_proxy({"autobus.example": "testobjectserver2"},
                multiple=True) as proxy:
            proxy.watch_object("test", changed)
            wait_for_interrupt()




from __future__ import with_statement
import os
import platform
import sys


# We add some special processing stuff if we're on Jython to get things to
# work correctly. Additionally, the Java port of libautobus uses the Python
# version running under Jython; to make its job easier, we'll also do some
# special processing if Java's libautobus is on the classpath. is_jython is
# set to true if we're under Jython, and is_java_lib is set to true if Java's
# libautobus is on the classpath.
is_jython = platform.system() == "Java"
try:
    import afn.libautobus.InterfaceWrapper as iw_temp #@UnresolvedImport
    del iw_temp
    is_java_lib = True
except:
    is_java_lib = False

__doc__ = """\
This module is the Python client library for Autobus. It's the authoritative
Autobus client library; features generally appear here first and then get
ported to versions of libautobus written for other languages.

To use this module, you'll generally start with an instance of
AutobusConnection. See that class's documentation for more information on how
to use it.

This module includes some hacks to get it to work properly on Jython. In
particular, Google's Protocol Buffers, which libautobus uses, do not work
properly on Jython by default, but simply importing libautobus adds a fix to
get it to work. (In fact, other Jython applications that want to make use of
Protocol Buffers could import libautobus just to get protocol buffers working.)
However, there are some issues right now, namely that if you're going to be
using libautobus on Jython, you need to have the Java libautobus on your
classpath. The reason is that this module changes some of its classes to
implement interfaces in the Java version, and the Java version then makes use
of this module running under Jython. As long as the Java libautobus is on the
classpath, you can import and use this module successfully from Jython code.
"""

# Workaround to get libautobus to work on Jython. Jython does not have Python's
# built-in buffer function, which Google's protocol buffers stuff needs to run.
# If we're on Jython, then, we'll create our own equivalent of the buffer
# function and install it in builtins.
if is_jython:
    def jython_buffer(object, offset=None, size=None):
        if size is not None:
            return object[offset:offset + size]
        elif offset is not None:
            return object[offset:]
        return object[:]
    sys.builtins["buffer"] = jython_buffer #@UndefinedVariable

from threading import Thread, RLock
from struct import pack, unpack
try:
    from json import dumps, loads
except ImportError:
    from simplejson import dumps, loads
# import autobus_protobuf.autobus_pb2 as protobuf
from traceback import print_exc
from socket import error as SocketError, socket as Socket, SHUT_RDWR
from functools import update_wrapper
from Queue import Queue, Empty
from time import sleep
import inspect
from datetime import datetime
from message_types import *
import atexit
if is_java_lib:
    from java.util import Map, List #@UnresolvedImport
    from org.python.core import Options #@UnresolvedImport
    from array import array
    from java.lang import System #@UnresolvedImport
    from afn.libautobus import (AutobusConnection as AutobusConnectionSuper, #@UnresolvedImport
            InterfaceWrapper as InterfaceWrapperSuper, #@UnresolvedImport
            FunctionWrapper as FunctionWrapperSuper, #@UnresolvedImport
            EventWrapper as EventWrapperSuper, #@UnresolvedImport
            ObjectWrapper as ObjectWrapperSuper) #@UnresolvedImport
    Options.respectJavaAccessibility = False

# We want to allow some additional types beyond the normal Python types to be
# encoded, so we'll use a function to do the translation
if is_java_lib:
    def jython_encode_replace(object):
        if isinstance(object, Map):
            return dict(object)
        if isinstance(object, (List, array)):
            return list(object)
        return object
else:
    def jython_encode_replace(object):
        return object
if not is_java_lib:
    AutobusConnectionSuper = object
    InterfaceWrapperSuper = object
    EventWrapperSuper = object # test file
    FunctionWrapperSuper = object
    ObjectWrapperSuper = object

COMMAND = 1
RESPONSE = 2
NOTIFICATION = 3

DEFAULT_PORT = 28862

MESSAGE_SEPARATORS = ",", ":"

next_id = 1
next_id_lock = RLock()

def get_next_id():
    """
    Thread-safely increments next_id and returns the value it previously held.
    """
    global next_id
    with next_id_lock:
        the_id = next_id
        next_id += 1
    return the_id

live_connections = []

@atexit.register
def on_exit():
    for connection in live_connections[:]:
        try:
            connection.shutdown()
        except:
            pass

def create_message(action, message_type=COMMAND, **kwargs):
    """
    Creates a new dictionary representing a message. The dict's action key will
    be set to the specified action, which should be one of the constants
    defined in libautobus.message_types. If the specified message type is an
    integer (as COMMAND, RESPONSE, and NOTIFICATION are), it is set as the
    dict's message_type attribute, and the dict's message_id attribute is set
    to a newly-generated id. Otherwise, message_type is assumed to be a dict
    representing a message that this one should be constructed in reply to, and
    this dict's message_type will be set to RESPONSE and its message_id will
    be set to the specified message's message_id. The new dict is then
    returned.
    
    One additional thing I added: if message_type is a message to reply to,
    but its message_type is NOTIFICATION, the returned dict will have its
    message_type set to None. Connection.send ignores such messages, which will
    properly result in a response not being sent for a notification without
    whatever code that may be using this function having any idea about it.
    
    Any additional keyword parameters to this function are set as attributes
    on the returned dict. 
    """
    message = {"action": action}
    if isinstance(message_type, dict):
        message["message_id"] = message_type["message_id"]
        if message_type["message_type"] == NOTIFICATION:
            message["message_type"] = None
        else:
            message["message_type"] = RESPONSE
    else:
        message["message_id"] = get_next_id()
        message["message_type"] = message_type
    message.update(kwargs)
    return message

def read_fully(socket, length):
    """
    Reads the specified number of bytes from the socket. The socket will be
    repeatedly read until this many bytes have been received. They will then
    be returned as a string. If the socket dies before this many bytes can be
    read, EOFError will be raised.
    """
    data = ""
    while len(data) < length:
#        print "Reading bytes..."
        try:
            new_data = socket.recv(length - len(data))
        except: # Sporadically happens if the socket dies
            new_data = ""
#        print str(len(new_data)) + " bytes received"
        if new_data == "":
            raise EOFError()
        data += new_data
    return data

def linesplit(socket):
    """
    A generator that yields lines of text from the socket.
    """
    buffer = socket.recv(512)
    done = False
    while not done:
        if "\n" in buffer:
            line, buffer = buffer.split("\n", 1)
            if line[-1:] == "\r":
                line = line[:-1]
            if line.strip(): # Ignore blank lines
                yield line
        else:
            more = socket.recv(512)
            if not more:
                done = True
            else:
                buffer = buffer + more
    if buffer[-1:] == "\r":
        buffer = buffer[:-1]
    if buffer:
        yield buffer

class InputThread(Thread):
    """
    A thread that reads messages from a socket and calls a user-defined
    function with the received messages. When the connection dies, it closes
    the socket and calls another user-defined function.
    """
    def __init__(self, socket, message_function, close_function=None):
        Thread.__init__(self)
        self.socket = socket
        self.generator = linesplit(socket)
        self.message_function = message_function
        self.close_function = close_function
    
    def run(self):
        try:
            while True:
                message_data = self.generator.next()
#                print "Message received, decoding..."
                message = loads(message_data)
#                print "Dispatching received message..."
                self.message_function(message)
        except EOFError:
            pass
        except StopIteration:
            pass
        except SocketError:
            pass
        except:
            print_exc()
#        print "Input thread finished, closing socket..."
        self.socket.close()
        if self.close_function:
#            print "Invoking close_function..."
            self.close_function()
#        print "Input thread has shut down."


class OutputThread(Thread):
    """
    A thread that repeatedly calls a function which should return a message
    object to send. The function can (and generally should) block as needed.
    This thread then writes the message to the socket. When the socket dies,
    it will be closed. If the function returns None, the socket will be
    immediately closed and this thread will exit.
    
    finished_function will be called after every message has been successfully
    written out. It defaults to lambda: None.
    
    The read function can return either a Message object or a string. If it's
    the former, it will be converted to a string via the message's
    SerializeToString function. Messages are ordinarily returned as Message
    objects to prevent message serialization from blocking the event queue in
    Autobus, but in some cases (such as when a single message is to be sent to
    several recipients), it may be more efficient to pre-encode the message
    and send it as a string.
    """
    def __init__(self, socket, read_function, finished_function=lambda: None,
            shut_on_end=False):
        Thread.__init__(self)
        self.socket = socket
        self.read_function = read_function
        self.finished_function = finished_function
        self.enable_buffer_hack = False
        self.shut_on_end=shut_on_end
    
    def run(self):
        try:
            while True:
#                print "Waiting for a message to send..."
                message = self.read_function()
                try:
#                print "Getting ready to send message " + str(message)
                    if message is None:
#                        print "Message was empty, shutting down the thread"
                        break
                    if isinstance(message, str):
                        message_data = message
                    else:
                        message_data = dumps(message, separators=MESSAGE_SEPARATORS)
                    self.socket.sendall(message_data + "\r\n")
                    if self.enable_buffer_hack:
                        self.socket.sendall((" "*5120) + "\r\n")
                finally:
                    self.finished_function()
        except SocketError:
            pass
        except:
            print_exc()
        if self.shut_on_end:
            try:
                self.socket.shutdown(SHUT_RDWR)
            except:
                pass
        try:
            self.socket.close()
        except:
            pass

class NoSuchInterfaceException(Exception):
    def __init__(self, interface_name, info=None):
        self.interface_name = interface_name
        self.info = info
    
    def __str__(self):
        value = "No such interface: " + self.interface_name
        if self.info is not None:
            value += ", additional info: " + self.info
        return value

class NoSuchFunctionException(Exception):
    def __init__(self, function_name):
        self.function_name = function_name
    
    def __str__(self):
        return "No such function: " + self.function_name

class TimeoutException(Exception):
    """
    An exception thrown from within function calls and some other places when
    the server or the remote client didn't respond within a reasonable time
    (currently 30 seconds).
    """

class NotConnectedException(Exception):
    """
    An exception thrown from within function calls and some other places when
    there is not currently a connection to the Autobus server.
    """

def discard_args(function):
    """
    Returns a function that calls the specified function, discarding all
    arguments passed into it. The specified function is invoked with no
    arguments.
    """
    def wrapper(*args, **kwargs):
        function()
    update_wrapper(wrapper, function)
    return wrapper

def merge_repeated(src, dest):
    """
    Merges all items in the repeated field src into the repeated field dest. 
    """
    for item in src:
        dest.add().MergeFrom(item)

def encode_object(object):
    """
    Encodes an object to its json-encoded form.
    """
    object = jython_encode_replace(object)
    if isinstance(object, (bool, int, long, float, basestring)):
        instance = object
    elif isinstance(object, datetime):
        instance = ["timestamp", {"year": object.year,
                "month": object.month, "day": object.day, "hour": object.hour,
                "minute": object.minute, "second": object.second,
                "millisecond": object.microsecond / 1000}]
    elif object is None:
        instance = ["null", None]
    elif isinstance(object, (list, tuple)):
        instance = ["list", [encode_object(o) for o in object]]
    elif isinstance(object, dict):
        keys = []
        values = []
        for key, value in object.items():
            keys.append(encode_object(key))
            values.append(encode_object(value))
        instance = ["map", [keys, values]]
    elif isinstance(object, Exception):
        instance = ["exception", {"text": type(object).__name__ + ": " + str(object)}]
    else:
        raise Exception("Invalid instance type to encode: " + 
                str(type(object)))
    return instance

def decode_object(instance):
    if isinstance(instance, (bool, int, long, float, basestring)):
        return instance
    # This is a custom type packed as a list with two items, so we'll unpack it
    instance_type, value = instance;
    if instance_type == "timestamp":
        return datetime(year=value["year"], month=value["month"],
                day=value["day"], hour=value["hour"],
                minute=value["minute"], second=value["second"],
                microsecond=value["millisecond"] * 1000)
    elif instance_type == "null":
        return None
    elif instance_type == "list":
        return [decode_object(o) for o in value]
    elif instance_type == "map":
        result = {}
        keys, values = value
        for key, value in zip(keys, values):
            result[decode_object(key)] = decode_object(value)
        return result
    elif instance_type == "exception":
        return Exception(value["text"])
    else:
        raise Exception("Invalid instance type to decode: " + 
                str(instance_type))

def get_function_doc(interface, function_name, remove_self=True):
    function = getattr(interface, function_name)
    is_method_before = inspect.ismethod(function)
    while hasattr(function, "wrapped"):
        function = function.wrapped
    if is_method_before or inspect.ismethod(function):
        argspec = inspect.getargspec(function)
        # Remove the leading "self" argument
        if remove_self:
            argspec = (argspec[0][1:],) + argspec[1:]
        args = inspect.formatargspec(*argspec)
    elif inspect.isfunction(function):
        args = inspect.formatargspec(*inspect.getargspec(function))
    else:
        args = "(...)"
    doc = inspect.getdoc(function)
    if doc is None:
        doc = ""
    return function_name + args + "\n\n" + doc

def start_thread(function):
    """
    A function decorator that can be used on functions registered to
    interfaces. Such functions will be invoked in a new thread, but their
    return value will still be captured and sent back to the client that
    invoked the function. This allows the function to make additional autobus
    calls while it's running and return results based on those calls, but no
    guarantee is made that two invocations of the function will not run at the
    same time.
    
    The same effect can be achieved by setting the attribute start_thread to
    have the value True on the function object. Indeed, this is exactly what
    this decorator does.
    """
    function.start_thread = True
    return function

class InterfaceWrapper(InterfaceWrapperSuper):
    def __init__(self, connection, name):
        self.connection = connection
        self.name = name
    
    def __getattr__(self, attribute):
        return FunctionWrapper(self.connection, self, attribute)
    
    def __getitem__(self, item):
        if isinstance(item, slice):
            if slice.start == "object":
                return ObjectWrapper(self, item.stop)
            raise ValueError("Invalid type of proxy to get: " + slice.stop)
        if isinstance(item, basestring):
            return self.__getattr__(item)
        raise TypeError("Invalid item type: " + str(type(item)))
    
    if is_java_lib:
        get_function = __getattr__
    
    def __str__(self):
        return "<InterfaceWrapper on interface " + self.name + ">"

class FunctionWrapper(FunctionWrapperSuper):
    def __init__(self, connection, interface, name):
        self.connection = connection
        self.interface = interface
        self.name = name
    
    def __call__(self, *args):
        instance_args = [encode_object(arg) for arg in args]
        message = create_message(CallFunctionCommand,
                interface_name=self.interface.name, function=self.name,
                arguments=instance_args)
        response = self.connection.query(message)
        if response["action"] == ErrorResponse:
            raise Exception("Server-side error: " + response.get("text",
                    "No additional error information was sent by the server."))
        result = decode_object(response["return_value"])
        if isinstance(result, Exception):
            raise result
        return result
    
    def invoke(self, args):
        """
        Calls this function, passing in the arguments in the specified sequence
        object. This is currently equivalent to self(*args), and is present
        mostly to make using libautobus from Jython easier.
        """
        return self(*args)

    def invoke_py(self, args):
        """
        Exactly the same as invoke(args). The Java version of libautobus that
        wraps this function returns a PyObject instead of an Object, which
        changes some semantics slightly. In other words, this function exists
        purely for the benefit of the Java Autobus API.
        """
        return self(*args)
    
    def send(self, *args):
        """
        Invokes this function asynchronously. self(...) is exactly the same as
        self.send(...), except that the call is sent off to the server, None
        is returned to the caller, and the return value or exception send from
        the server, if any, is discarded.
        """
        instance_args = [encode_object(arg) for arg in args]
        message = create_message(CallFunctionCommand,
                NOTIFICATION, interface_name=self.interface.name,
                function=self.name, arguments=instance_args)
        self.connection.send(message)
    
    def invoke_later(self, args):
        """
        Same as send(*args). This exists to make libautobus from Java easier.
        """
        return self.send(*args)
    
    def invoke_later_py(self, args):
        return self.invoke_later(args)
    
    def __str__(self):
        return ("<FunctionWrapper on function " + self.name + " and interface "
                + str(self.interface) + ">")
    
    __repr__ = __str__

class ObjectWrapper(ObjectWrapperSuper):
    def __init__(self, interface, name):
        self.interface = interface
        self.name = name
    
    def watch(self, function):
        """
        Exactly the same as calling add_object_watch on the connection for
        this object and the specified function. In fact, this function just
        delegates to add_object_watch at present.
        """     
        self.interface.connection.add_object_watch(self.interface.name,
                self.name, function)   
    
    def __str__(self):
        return ("<ObjectWrapper on object " + self.name + " and interface "
                + str(self.interface) + ">")
    __repr__ = __str__

class EventWrapper(EventWrapperSuper):
    pass

class LocalInterface(object):
    """
    A local interface. Instances of this class represent interfaces that have
    been added locally to the AutobusConnection instance. They will be
    registered with the server when a connection is established.
    """
    def __init__(self, name, doc):
        self.connection = None
        self.name = name
        self.doc = doc
        if doc is None:
            self.doc = ""
        self.functions = {} # Maps names to LocalFunction objects representing
        # functions we've registered on this interface
        self.events = {} # Maps names to LocalEvent objects representing
        # events we've registered on this interface
        self.objects = {} # Maps names to LocalObject objects representing
        # objects we've registered on this interface
    
    def add_all(self, interface):
        """
        Adds all of the functions present on the specified object that do not
        start with an underscore to this interface. This is used by
        AutobusConnection.add_interface.
        """
        for function_name in dir(interface):
            if function_name[0:1] != "_" and callable(
                    getattr(interface, function_name)):
                self.functions[function_name] = LocalFunction(function_name,
                        get_function_doc(interface, function_name),
                        getattr(interface, function_name))
    def register_function(self, local_function):
        """
        Registers the specified LocalFunction object to this interface.
        """
        self.functions[local_function.name] = local_function
        local_function.interface = self
    
    def register_event(self, local_event):
        """
        Registers the specified LocalEvent object to this interface.
        """
        self.events[local_event.name] = local_event
        local_event.interface = self
    
    def register_object(self, local_object):
        """
        Registers the specified LocalObject object to this interface.
        """
        self.objects[local_object.name] = local_object
        local_object.interface = self

class LocalFunction(object):
    """
    A local function. Local functions are stored in a LocalInterface instance's
    function map.
    """
    def __init__(self, name, doc, function):
        self.interface = None
        self.name = name
        self.doc = doc
        if doc is None:
            self.doc = ""
        if is_java_lib and hasattr(function, "run"):
            self.start_thread = function.isInNewThread()
            self.function = function.run
        else:
            self.start_thread = hasattr(function, "start_thread") and function.start_thread
            self.function = function

class LocalEvent(object):
    """
    A local event. Local events are callable; calling one will fire it.
    """
    def __init__(self, name, doc):
        """
        Creates a new local event. The event will use the specified name and
        docstring.
        """
        self.interface = None
        self.name = name
        self.doc = doc
    
    def __call__(self, *arguments):
        """
        Fires this event, using the specified arguments as the event's
        arguments. If no connection to the server is available, this does
        nothing.
        """
        with self.interface.connection.on_connect_lock:
            message = create_message(FireEventCommand,
                    NOTIFICATION, interface_name=self.interface.name,
                    event_name=self.name,
                    arguments=[encode_object(o) for o in arguments])
            try:
                self.interface.connection.send(message)
            except NotConnectedException: # We don't have a connection to the
                # server right now, so we'll just ignore the attempt to fire
                # the event.
                pass

class LocalObject(object):
    """
    A local object.
    """
    def __init__(self, name, doc, value):
        """
        Creates a new local object. The object will use the specified name,
        docstring, and it will have the specified initial value.
        """
        self.interface = None
        self.name = name
        self.doc = doc
        if doc is None:
            self.doc = ""
        encode_object(value) # Make sure in advance that we can encode the
        # value the user's trying to put as the object's default value
        self.value = value
    
    def set(self, new_value):
        encode_object(new_value) # Same thing as in the constructor; make sure
        # in advance we can successfully encode the new value for the object
        self.value = new_value
        self.notify()
    
    def notify(self):
        """
        Sends a message to the server notifying it that this object's value has
        changed. The encoded form of the object will be sent in the message.
        """
        with self.interface.connection.on_connect_lock:
            message = create_message(SetObjectCommand,
                    NOTIFICATION, interface_name=self.interface.name,
                    object_name=self.name, value=encode_object(self.value))
            try:
                self.interface.connection.send(message)
            except NotConnectedException: # We've already stored the value
                # locally if we're calling notify, so we'll just rely on the
                # connect thread to send this value on next connection.
                pass


class AutobusConnection(AutobusConnectionSuper):
    """
    A connection to an Autobus server. The typical way to use libautobus is to
    create an instance of this class and go from there.
    
    Right now, all interfaces that are to be made available must be specified
    before the connection is connected for the first time. They can be
    specified after, but this class won't register them until it reconnects.
    
    If the connection to the autobus server is lost, all currently-pending
    functions etc will raise an exception, and this class will attempt to
    re-establish a connection and re-register interfaces.
    
    Subscripting an instance of this class is the same as calling
    get_interface(). For example, a wrapper around the interface "example" on
    the server could be obtained with some_autobus_connection["example"].
    
    Autobus provides introspection and additional information via the built-in
    autobus interface. You can access it with get_interface("autobus") or
    connection["autobus"]. Calling
    connection["autobus"].list_functions("autobus") will list all of the
    functions available on the autobus interface along with more documentation
    on how to use them.
    
    If reconnect is True (the default), the connection will reconnect itself
    and re-register all of its local interfaces and functions when it gets
    disconnected. It will continue attempting to reconnect indefinitely. If
    reconnect is False, it's up to the on_disconnect function (or some other
    functionality) to call the connect method when the connection is lost.
    
    If print_exceptions is True, any exceptions thrown by a local function
    invoked by a remote client will be printed with traceback.print_exc()
    before being sent back to the client that invoked the function. This is
    useful when a function is raising an unexpected error and more information
    about the error, such as its traceback, is needed.
    
    Regardless of whether or not you eventually connect an instance of this
    class, you should call its shutdown function when you're done with it.
    If you don't, a memory leak will result. The reason for this is that
    libautobus tracks connections that have not yet been shut down in a list so
    that they can be automatically shut down when Python exits. Of course, if
    you're going to create a single long-lived connection to serve from a
    particular interface, that's fine. But if you're going to be creating
    several connections over the lifetime of the application, make sure you
    shut them down when you're done with them. 
    """
    def __init__(self, host=None, port=None, reconnect=True,
            print_exceptions=False, on_connect=lambda: None,
            on_disconnect=lambda: None):
        """
        Creates a new connection. This doesn't actually connect to the
        autobus server; use connect() or start_connecting() for that.
        """
        if host is None:
            host = os.getenv("AUTOBUS_SERVER")
        if host is None and is_java_lib:
            host = System.getProperty("autobus.server")
        if host is None:
            host = "localhost"
        if port is None:
            port = os.getenv("AUTOBUS_PORT")
        if port is None:
            port = DEFAULT_PORT
        if isinstance(port, basestring):
            port = int(port)
        self.socket = None
        self.host = host
        self.port = port
        self.reconnect = reconnect
        self.on_connect = on_connect
        self.on_disconnect = on_disconnect
        self.print_exceptions = print_exceptions
        self.on_connect_lock = RLock()
        self.input_thread = None
        self.output_thread = None
        self.interfaces = {} # Map of names to LocalInterface objects
        # representing interfaces we've registered
        self.is_shut_down = False
        self.send_queue = Queue()
        self.receive_queues = {} # Maps message ids expecting responses to the
        # corresponding queues waiting for the message response
        self.event_listeners = {} # Maps tuples containing an interface name
        # and an event name to a list of functions listening for the specified
        # event to fire on the server. At least one entry must be present in
        # each list in order for AutobusConnection to auto-register a
        # listener on the specified event on connect, even if it's as simple
        # as lambda: None.
        self.object_values = {} # Maps tuples containing an interface name and
        # an object name to the object's current value as sent by the server.
        # This dict gets replaced every time we disconnect.
        self.object_listeners = {} # Maps tuples containing an interface name
        # and an object name to a list of functions listening for changes in
        # that object. At least one entry must be present in each list in order
        # for AutobusConnection to auto-register a watch on the object on
        # connect, even if it's as simple as lambda: None.
        live_connections.append(self)
    
    def shutdown(self):
        self.is_shut_down = True
        try:
            self.socket.shutdown(SHUT_RDWR)
        except:
            pass
        try:
            self.socket.close()
        except:
            pass
        live_connections.remove(self)
    
    def add_interface(self, name, interface=None):
        """
        Adds an interface that will be automatically registered with the server
        on connecting. All methods that do not start with an underscore on the
        specified object will be registered on the interface as functions. The
        specified object's docstring, if it has one, will be used as the
        interface's documentation.
        
        If the specified interface is None, no functions will be registered on
        the interface. Either way, functions can later be registered with
        add_function.
        """
        self.interfaces[name] = LocalInterface(name, inspect.getdoc(interface))
        self.interfaces[name].connection = self
        if interface is not None:
            self.interfaces[name].add_all(interface)
    
    def add_function(self, interface_name, function_name, doc, function):
        """
        Adds a function to the specified local interface. The interface must
        already have been registered with add_interface. The function will use
        the specified name and documentation (which must not be None: it should
        be the empty string if no documentation is available). When another
        client calls the specified function, the last argument to this
        function, which should be callable, will be invoked and its return
        value sent back to the invoking client. The function can raise an
        exception, which will be re-thrown on the client side.
        """
        self.interfaces[interface_name].register_function(LocalFunction(
                function_name, doc, function))
    
    def add_event_listener(self, interface_name, event_name, function):
        """
        Adds a function that will be notified when the specified remote event
        is fired. This function is not thread-safe and should generally only
        be called before the first connection to the server is created.
        
        The specified function should accept a number of arguments
        corresponding to the number of arguments sent when the remote client
        fires the event. You'll usually have to consult the event's
        documentation.
        """
        if hasattr(function, "fired"): # We're on Jython and the function is
            # an instance of EventListener
            function = function.fired
        event_spec = interface_name, event_name
        if event_spec not in self.event_listeners:
            self.event_listeners[event_spec] = []
        self.event_listeners[event_spec].append(function)
    
    def add_event(self, interface_name, event_name, doc):
        """
        Adds a local event to the specified interface. The LocalEvent
        instance created for the event will be returned.
        """
        event = LocalEvent(event_name, doc)
        self.interfaces[interface_name].register_event(event)
        return event
    
    def add_object_watch(self, interface_name, object_name, function):
        """
        Adds a function that will be notified when the specified remote
        object's value changes. add_object_watch is not thread-safe and should
        generally only be called before the first connection to the server is
        created.
        
        The function should accept one argument. This argument will be the new
        value of the object.
        
        The function must not interact with Autobus. If it does, deadlock will
        likely result, as the input thread, which receives packets from the
        server (such as responses to function calls), blocks while this
        function is invoked. If you need to interact with Autobus (for example,
        to call functions etc), the function you supply should start a new
        thread to do so.
        """
        if hasattr(function, "changed"): # We're on Jython and the function
            # is an instance of ObjectListener
            function = function.changed
        object_spec = interface_name, object_name
        if object_spec not in self.object_listeners:
            self.object_listeners[object_spec] = []
        self.object_listeners[object_spec].append(function)
    
    def add_object(self, interface_name, object_name, doc, value):
        """
        Adds a shared object to the specified interface. The LocalObject
        instance created for the object will be returned.
        """
        object = LocalObject(object_name, doc, value)
        self.interfaces[interface_name].register_object(object)
        return object
    
    def start_connecting(self):
        """
        Calls connect(None) in a new thread. Any exception thrown by the call
        to connect(None) is discarded without being printed to stdout.
        """
        def target():
            try:
                self.connect(attempts=None)
            except:
                pass
        Thread(target=target).start()
    
    def connect(self, attempts=1):
        """
        Connects to the Autobus server. The specified number of attempts will
        be made to re-establish a connection, each time waiting an amount of
        time that increments itself up to 20 seconds. If attempts is None, an
        infinite number of attempts will be made. Once a connection has been
        established, this method returns. If it fails to establish a
        connection, an exception will be raised.
        """
        if attempts == 0:
            attempts = None
        progress = 0
        delay = 0.1
        delay_increment = 1.5
        while (attempts is None or progress < attempts) and not self.is_shut_down:
            progress += 1
            delay *= delay_increment
            if delay > 20:
                delay = 20
            self.socket = Socket()
            try:
                self.socket.connect((self.host, self.port))
            except:
                sleep(delay)
                continue
            with self.on_connect_lock:
                self.connection_established()
            return
        raise NotConnectedException("Couldn't connect")
    # messagearrived, inputclosed, readnextmessage
    
    def connection_established(self):
#        print "Processing established connection..."
        self.input_thread = InputThread(self.socket, self.message_arrived, self.input_closed)
        self.output_thread = OutputThread(self.socket, self.read_next_message, self.message_write_finished)
        self.send_queue = Queue() # clear the send queue
        self.receive_queues = {}
        self.input_thread.start()
        self.output_thread.start()
#        print "Registering interfaces with the server..."
        for interface in self.interfaces.values():
            self.send(create_message(RegisterInterfaceCommand,
                    NOTIFICATION, name=interface.name, doc=interface.doc))
            message = create_message(RegisterFunctionCommand,
                    NOTIFICATION, interface_name=interface.name, name=[], doc=[])
            for function in interface.functions.values():
                message["name"].append(function.name)
                message["doc"].append(function.doc)
            self.send(message)
            for event in interface.events.values():
                self.send(create_message(
                        RegisterEventCommand, NOTIFICATION,
                        interface_name=interface.name, event_name=event.name,
                        doc=event.doc))
            for object in interface.objects.values():
                self.send(create_message(
                        RegisterObjectCommand, NOTIFICATION,
                        interface_name=interface.name, object_name=object.name,
                        doc=object.doc, value=encode_object(object.value)))
        for interface_name, event_name in self.event_listeners.keys():
            self.send(create_message(
                    RegisterListenerCommand, NOTIFICATION,
                    interface_name=interface_name, event_name=event_name))
        for interface_name, object_name in self.object_listeners.keys():
            message = create_message(
                    WatchObjectCommand, COMMAND,
                    interface_name=interface_name, object_name=object_name)
            self.send(message) # I used to send this as a query and set the
            # resulting value into the object, but this caused a race condition
            # where another value arriving as a SetObjectCommand before the
            # logic in this method could get the new value set into the object
            # would cause the object to be reset to its initial value. So for
            # now, we're processing the WatchObjectResponse in message_arrived
            # the same as we process SetObjectCommand instances from the
            # server, which fixes the race condition. That's why it's a command
            # but we're not querying for the response.
        # That's it! Now we run the user's on_connect function.
        self.on_connect()
        
    def message_arrived(self, message):
#        print "Message from the server arrived: " + str((message)) + "\n\n"
        queue = self.receive_queues.get(message["message_id"], None)
        if queue is not None:
            queue.put(message)
            try:
                del self.receive_queues[message["message_id"]]
            except KeyError:
                pass
            return
        if message["action"] == RunFunctionCommand:
            interface_name = message["interface_name"]
            function_name = message["function"]
            arguments = [decode_object(arg) for arg in message["arguments"]]
            try:
                if function_name.startswith("_"):
                    raise Exception("Function names starting with _ are "
                            "not allowed")
                interface = self.interfaces[interface_name]
                function = interface.functions[function_name]
            except Exception, e:
                if self.print_exceptions:
                    print_exc()
                self.send(create_message(RunFunctionResponse,
                        message, return_value=encode_object(e)))
            def process():
                try:
                    return_value = function.function(*arguments)
                except:
                    if self.print_exceptions:
                        print_exc()
                    _, return_value, _ = sys.exc_info() #@UndefinedVariable
                self.send(create_message(RunFunctionResponse,
                        message, return_value=encode_object(return_value)))
            if function.start_thread:
                Thread(target=process).start()
            else:
                process()
            return
        if message["action"] == FireEventCommand:
            interface_name = message["interface_name"]
            event_name = message["event_name"]
            arguments = message["arguments"]
            event_spec = (interface_name, event_name)
            self.fire_event_listeners(event_spec, [decode_object(o) for o in arguments])
            return
        if message["action"] in (SetObjectCommand, WatchObjectResponse):
            interface_name = message["interface_name"]
            object_name = message["object_name"]
            value = message["value"]
            object_spec = (interface_name, object_name)
            self.object_values[object_spec] = decode_object(value)
            self.notify_object_listeners(object_spec)
            return
        if message["action"] == PingCommand:
            self.send(create_message(PingResponse, message))
            return
        print "Not processing message for action " + message["action"]
    
    def notify_object_listeners(self, object_spec):
        """
        This should only be used by AutobusConnection itself. Notifies all
        listeners on the specified object spec about the object's current
        value. 
        """
        value = self.object_values.get(object_spec, None)
        if object_spec in self.object_listeners:
            for listener in self.object_listeners[object_spec]:
                try:
                    listener(value)
                except:
                    print_exc()
    
    def fire_event_listeners(self, event_spec, arguments):
        """
        This should only be used by AutobusConnection itself. Notifies all
        listeners on the specified event spec that the event has been fired
        with the specified arguments.
        """
        if event_spec in self.event_listeners:
            for listener in self.event_listeners[event_spec]:
                try:
                    listener(*arguments)
                except:
                    print_exc()
    
    def query(self, message, timeout=30):
        """
        Sends the specified message from the server, waiting up to the
        specified timeout for a response. When a response is received, it is
        returned. If a response is not received within the specified amount of
        time, a TimeoutException will be raised. Otherwise, the response sent
        by the server will be returned. 
        """
        queue = Queue()
        if self.receive_queues is None or self.input_thread is None:
            raise NotConnectedException()
        self.receive_queues[message["message_id"]] = queue
        self.send(message)
        try:
            response = queue.get(block=True, timeout=timeout)
        except Empty:
            try:
                del self.receive_queues[message["message_id"]]
            except KeyError:
                pass
            raise TimeoutException()
        # If we don't get a KeyError, it's guaranteed that the input thread
        # will already have removed the queue, so we don't need to worry about
        # removing it.
        return response
    
    def send(self, message):
        """
        Adds the specified message, which should be an instance of
        protobuf.Message, to the send queue. If the message is None, this
        method returns without doing anything.
        """
        if message is not None and message["message_type"] is not None:
#            print "Adding message to send queue " + str(self.send_queue) + "..."
            queue = self.send_queue
            if queue is None:
                raise NotConnectedException()
            queue.put(message, block=True)
    
    def input_closed(self):
        self.send_queue.put(None)
        try:
            while True:
                self.send_queue.task_done()
        except ValueError: # Happens once we've marked all tasks as complete
            pass
        for queue in self.receive_queues.values():
            queue.put(None)
        self.receive_queues.clear()
        self.input_thread = None
        self.output_thread = None
        self.send_queue = None
        self.receive_queues = None
        self.object_values.clear()
        self.object_values = {}
        for object_spec in self.object_listeners.keys():
            self.notify_object_listeners(object_spec)
        self.on_disconnect()
        if self.reconnect and not self.is_shut_down:
            self.start_connecting()
    
    def read_next_message(self):
        send_queue = self.send_queue
        if send_queue is None:
            return None
#        print "Getting next message from send queue " + str(send_queue) + "..."
        return send_queue.get(block=True)
    
    def message_write_finished(self):
        send_queue = self.send_queue
        if send_queue is not None:
            try:
                send_queue.task_done()
            except:
                pass
    
    def send_error(self, in_reply_to=None, **kwargs):
        """
        Sends an error back to the client. If in_reply_to is itself a
        notification message, no message will be sent back. Otherwise, an
        ErrorResponse to the specified message will be sent. The error's
        fields will be initialized to have any additional keyword arugments
        specified. You'll typically do:
        
        send_error(some_message, text="A random problem happened")
        """
        if in_reply_to == None:
            in_reply_to = NOTIFICATION
        message = create_message(ErrorResponse, in_reply_to, **kwargs)
        self.send(message)
    
    def get_interface(self, name):
        """
        Returns a wrapper around the specified interface. This doesn't check
        to make sure that the specified interface is actually present on the
        server; if it's not, an exception will be thrown when a function on
        the interface is invoked for the first time.
        
        This function can be used like so:
        
        interface = some_connection.get_interface("example")
        return_value = interface.some_function("first", 2, ["third"])
        
        This would invoke the function "some_function" on the interface
        "example" on the server. An exception would be thrown if the specified
        interface does not exist on the server or if the specified interface
        does not have any such function.
        
        If the function does not return within 30 seconds (or the server does
        not send a response indicating that the function has returned within
        that time frame), a TimeoutException will be raised.
        
        This function returns a wrapper around a server-side interface. To get
        access to a local interface to register functions, events, or objects,
        use get_local_interface.
        """
        return InterfaceWrapper(self, name)
    
    __getitem__ = get_interface
    
    def get_local_event(self, interface_name, event_name):
        return self.interfaces[interface_name].events[event_name]
    
    def get_local_object(self, interface_name, object_name):
        """
        Returns the LocalObject instance representing the specified local
        object, which must have been previously registered with this
        connection. This is short for
        get_local_interface(interface_name).objects[object_name].
        """
        return self.interfaces[interface_name].objects[object_name]
    
    def get_local_function(self, interface_name, function_name):
        """
        Returns the LocalFunction instance representing the specified local
        function, which must have been previously registered with this
        connection. This is short for
        get_local_interface(interface_name).functions[function_name]
        """
        return self.interfaces[interface_name].functions[function_name]
    
    def get_local_interface(self, interface_name):
        """
        Returns the LocalInterface instance representing the specified local
        interface.
        """
        return self.interfaces[interface_name]
    
    def set_local_object(self, interface_name, object_name, value):
        """
        Sets the value of the specified local object and notifies the server if
        there's currently an active connection to the server. This is short for
        get_local_object(interface_name, object_name).set(value).
        """
        self.get_local_object(interface_name, object_name).set(value)
    
    def get_object_value(self, interface_name, object_name):
        """
        Returns the current value of a remote object, or None if the value is
        not known (which could be because add_object_watch was not called before
        the last time this connection successfully connected to the server).
        """
        return self.object_values.get((interface_name, object_name), None)
    
    def get_remote_object_value(self, interface_name, object_name):
        """
        Returns the current value of a remote object, or None if the value is
        not known. Unlike get_object_value, this function does not require the
        object to have been previously registered with add_object_watch, but
        it incurs a round-trip to the server to get the object's current
        value (even if the object has been registered with add_object_watch).
        Also unlike get_object_value, this function will raise
        NotConnectedException if there's no connection to the Autobus server
        or TimeoutException if the server did not respond within a timely
        manner (currently 30 seconds).
        """
        message = create_message(GetObjectValueCommand,
                interface_name=interface_name, object_name=object_name)
        response = self.query(message)
        if response["action"] == ErrorResponse:
            raise Exception("Server-side error: " + response.get("text",
                    "No additional error information was sent by the server."))
        return decode_object(response["value"])        
    
    def flush(self):
        """
        If this connection is currently connected, waits until all pending
        outbound messages have been successfully sent to the server.
        
        WARNING: I haven't gone through this with a fine-toothed comb for race
        conditions, so there very well could be some that could cause this
        function to deadlock. Use at your own risk, and thoroughly test any
        application using this (in particular, test out what happens when you
        kill the Autobus server around the time you call this method).
        """
        send_queue = self.send_queue
        if send_queue is not None:
            send_queue.join()
    
    def unshutdown(self):
        """
        Marks a shutdown Autobus connection as no longer shut down. This
        generally should not be used in production code; it's mostly for those
        that want to disconnect and reconnect an Autobus connection
        experimentally. The connection must have previously been shut down with
        shutdown() and sufficient time given for the connection to stop trying
        to reconnect. After this is called, all local interfaces, functions,
        objects, etc, will still be registered to this connection, and a call
        to connect() or start_connecting() immediately after calling this
        function will cause them to be re-registered with the server.
        """
        if self.is_shut_down:
            live_connections.append(self)
        self.is_shut_down = False
    
    def interrupt_loop(self):
        """
        Starts a loop waiting for a KeyboardInterrupt to occur. When one does,
        this bus will be shut down and this function will return.
        
        You don't need to call this function to use an Autobus connection at
        all since connections use threads to do all their useful work.
        It's intended primarily for daemon processes that exist solely to
        provide an Autobus interface and that would need to be notified when
        a Ctrl+C is used to kill the daemon.
        """
        try:
            while True:
                sleep(5)
        except KeyboardInterrupt:
            print "Interrupted, shutting down"
        finally:
            self.shutdown()

    




















from select import select
from Queue import Queue, Empty
from autobus2 import exceptions
from socket import socket as Socket, SHUT_RDWR, error as SocketError
from threading import Thread, RLock, Condition
from traceback import print_exc
import json
from utils import no_exceptions


class _InProgressSocketManager(object):
    """
    This class does not actually work right now. It's something that I'm
    writing that will some day replace InputThread and OutputThread and allow
    Autobus to function using only a single thread for each Bus instance. So
    it's a work in progress, and it doesn't actually work right now.
    """
    def __init__(self):
        # Set up the interrupt socket
        interrupt_server = Socket()
        interrupt_server.bind(("localhost", 0))
        interrupt_server.listen(1)
        self.interrupt_writer = Socket()
        self.interrupt_writer.setblocking(False)
        self.interrupt_writer.connect("localhost", interrupt_server.getsockname()[1])
        self.interrupt_reader = interrupt_server.accept()
        interrupt_server.shutdown(SHUT_RDWR)
        interrupt_server.close()
        self.interrupt_reader.setblocking(False)
        self.interrupt_writer.setblocking(False)
    
    def loop(self):
        pass
    
    def run_sync(self, function, timeout=None):
        q = Queue()
        @self.run_async
        def new_function():
            q.put(function())
        try:
            return q.get(timeout=timeout)
        except Empty:
            raise exceptions.TimeoutException
    
    def run_async(self, function):
        pass
    
    def interrupt(self):
        self.interrupt_writer.send(0)


def linesplit(socket):
    """
    A generator that yields lines of text from the socket.
    """
    buffer = socket.recv(512)
    done = False
    while not done:
        if "\n" in buffer:
            line, buffer = buffer.split("\n", 1)
            if line[-1:] == "\r":
                line = line[:-1]
            if line.strip(): # Ignore blank lines
                yield line
        else:
            more = socket.recv(512)
            if not more:
                done = True
            else:
                buffer = buffer + more
    if buffer[-1:] == "\r":
        buffer = buffer[:-1]
    if buffer:
        yield buffer


class InputThread(Thread):
    """
    A thread that reads messages from a socket and calls a user-defined
    function with the received messages. When the connection dies, it closes
    the socket and calls the function with None as its argument.
    
    The user-defined message function can be changed at any time by assigning
    to the function property of any given input thread. If None is assigned to
    it (or specified when constructing the input thread), the input thread will
    block after it receives a message until a non-None function is assigned to
    the function property.
    """
    def __init__(self, socket, function):
        Thread.__init__(self)
        self.socket = socket
        self.generator = linesplit(socket)
        self.lock = RLock()
        self.condition = Condition(self.lock)
        self.function = function
    
    @property
    def function(self):
        return self._function
    
    @function.setter
    def function(self, new_function):
        with self.lock:
            self._function = new_function
            self.condition.notify_all()
    
    def run(self):
        try:
            while True:
                message_data = self.generator.next()
#                print "Message received, decoding..."
                message = json.loads(message_data)
#                print "Dispatching received message..."
                f = self.get_function_or_wait()
                f(message)
        except EOFError:
            pass
        except StopIteration:
            pass
        except SocketError:
            pass
        except:
            print_exc()
#        print "Input thread finished, closing socket..."
        self.socket.close()
        f = self.get_function_or_wait()
        f(None)
    
    def get_function_or_wait(self):
        with self.lock:
            f = self.function
            while f is None:
                self.condition.wait()
                f = self.function
        return f


class OutputThread(Thread):
    """
    A thread that repeatedly calls a function which should return a message
    object to send. The function can (and generally should) block as needed.
    This thread then writes the message to the socket. When the socket dies,
    it will be closed. If the function returns None, the socket will be
    immediately closed and this thread will exit.
    
    finished_function will be called after every message has been successfully
    written out. It defaults to lambda: None.
    
    The read function can return either a json object or a string. If it's
    the former, it will be converted to a string by using the json module's
    json.dumps function. Messages are ordinarily returned as json
    objects to prevent message serialization from blocking the event queue in
    Autobus, but in some cases (such as when a single message is to be sent to
    several recipients), it may be more efficient to pre-encode the message
    and send it as a string.
    """
    def __init__(self, socket, read_function, finished_function=lambda: None,
            shut_on_end=False):
        Thread.__init__(self)
        self.socket = socket
        self.read_function = read_function
        self.finished_function = finished_function
        self.enable_buffer_hack = False
        self.shut_on_end=shut_on_end
    
    def run(self):
        try:
            while True:
#                print "Waiting for a message to send..."
                message = self.read_function()
                try:
#                print "Getting ready to send message " + str(message)
                    if message is None:
#                        print "Message was empty, shutting down the thread"
                        break
                    if isinstance(message, str):
                        message_data = message
                    else:
                        message_data = json.dumps(message)
                    self.socket.sendall(message_data + "\r\n")
                    if self.enable_buffer_hack:
                        self.socket.sendall((" "*5120) + "\r\n")
                finally:
                    self.finished_function()
        except SocketError:
            pass
        except:
            print_exc()
        if self.shut_on_end:
            try:
                self.socket.shutdown(SHUT_RDWR)
            except:
                pass
        try:
            self.socket.close()
        except:
            pass


def start_io_threads(socket, input_function, output_function):
    input_thread = InputThread(socket, input_function)
    output_thread = OutputThread(socket, output_function)
    input_thread.start()
    output_thread.start()
    return input_thread, output_thread


def shutdown(socket):
    with no_exceptions:
        socket.shutdown(SHUT_RDWR)
        socket.close()


def sendto(socket, data, address):
    """
    Sends the specified data, which should be a str, to the specified address,
    which is typically a (host, port) tuple, using the specified socket (which
    generally should be a UDP datagram socket). This basically just invokes
    socket.sendto(data, address), except that it catches and ignores any
    exceptions thrown while sending, which is useful because sending to
    255.255.255.255 when the network interface is down will cause an exception
    to be thrown.
    """
    with no_exceptions:
        socket.sendto(data, address)


def ensure_jsonable(value):
    try:
        json.dumps(value)
    except:
        raise exceptions.InvalidValueException()

































try:
    import argparse
except ImportError:
    from afn.backports import argparse


class AppendWithConst(argparse.Action):
    """
    An argparse action that functions like append, but it inserts the value of
    const as the first item in the list.
    """
    def __call__(self, parser, namespace, values, option_string=None):
        if values is None:
            values = []
        elif not isinstance(values, list):
            values = [values]
        items = argparse._ensure_value(namespace, self.dest, [])[:]
        items.append([self.const] + values)
        setattr(namespace, self.dest, items)




from collections import namedtuple

Answer = namedtuple("Answer", ["name", "type", "ttl", "items"])
A = namedtuple("A", ["ip"])
AAAA = namedtuple("AAAA", ["ip"])
CNAME = namedtuple("CNAME", ["target"])
MX = namedtuple("MX", ["ip"])



import gtk
import gtk.gdk
import gobject
import time
from threading import Thread
from functools import partial
from autobus2 import Bus
from afn.utils import Suppress

online_monitors = {}

def main():
    gobject.threads_init()
    with Bus() as bus:
        with bus.get_service_proxy({"type": "monitord"}, multiple=True) as proxy:
            w = gtk.Window()
            w.set_title("GMonitorOnline")
            vbox = gtk.VBox()
            w.add(vbox)
            vbox.show()
            label = gtk.Label()
            label.set_text("CPU usage:")
            vbox.add(label)
            label.show()
            w.show()
            def listener(connection, info, value):
                service_id = connection.service_id
                if value is None:
                    if service_id in online_monitors:
                        online_monitors[service_id].destroy()
                        del online_monitors[service_id]
                else:
                    if service_id not in online_monitors:
                        online_monitors[service_id] = gtk.ProgressBar()
                        online_monitors[service_id].set_text(info.get("hostname", "unknown"))
                        online_monitors[service_id].set_fraction(0)
                        vbox.add(online_monitors[service_id])
                        online_monitors[service_id].show()
                    # [0] is user, [1] is system
                    cpu = value["cpu"][0] + value["cpu"][1]
                    online_monitors[service_id].set_fraction(cpu / 100.0)
            proxy.watch_object("status", lambda * args: gobject.idle_add(partial(listener, *args)))
            gtk.main()










import pyaudio
import wave
import sys
import os.path
import time

if len(sys.argv) <= 2:
    print "You need to specify two arguments, the wordlist file to use and "
    print "the folder to record the voice into, respectively."
    sys.exit()


# FUNCTION DEFINITIONS
def compute_noise(the_block):
    the_noise = 0
    for i in xrange(0, len(the_block), 2):
        the_value = ord(the_block[i]) | (ord(the_block[i + 1]) << 8)
        if(the_value > 32767):
            the_value = the_value - 65536
        the_noise += the_value if the_value > 0 else the_value * -1
    return the_noise / (len(the_block) / 2)

def read_and_rotate():
    recent_blocks.append(mic.read(block_size))
    recent_noise.append(compute_noise(recent_blocks[-1]))
    if len(recent_blocks) > 10:
        del recent_blocks[0]
        del recent_noise[0]

def last_are_noisy(count):
    for i in xrange(1, count + 1):
        if recent_noise[-i] < noise_threshold:
            return False
    return True

def last_are_silent(count):
    for i in xrange(1, count + 1):
        if recent_noise[-i] > noise_threshold:
            return False
    return True

#def safe_word(text):
#    return re.sub("[^A-Za-z0-9]", "_", text)

# END OF FUNCTION DEFINITIONS

with open(sys.argv[1]) as wordlist_file:
    wordlist = [word.strip() for word in wordlist_file]
    wordlist_file.close()
wordlist = [word.lower() for word in wordlist if word[0] != "#"]
for word in wordlist:
    if wordlist.count(word) > 1:
        print "The wordlist specified contains the word " + repr(word) + " twice."
        sys.exit()

voice_folder = sys.argv[2]
voice_folder = voice_folder.replace("/", os.path.sep)

existing_words = [word[:-4].lower() for word in os.listdir(voice_folder) if word[-4:] == ".wav"]
new_words = [word for word in wordlist if word not in existing_words]
deleted_words = [word for word in existing_words if word not in wordlist]

print "There are " + str(len(new_words)) + " words for you to record: " + " ".join(new_words)
print ""
print str(len(deleted_words)) + " words are no longer in the wordlist: " + " ".join(deleted_words)
print ""
if len(new_words) == 0:
    print "You don't have any unrecorded words, so I'm going to exit. If you wanted "
    print "to re-record some words, delete them from your voice folder first, then "
    print "re-run this program."
    sys.exit()

print "Here's how it works: I'll tell you a word for you to record. NOTE: 0 should "
print "be pronounced as 'zero', not 'oh', and 'am' should be pronounced as 'AY EM', "
print "not 'am', and likewise for 'pm'. Once I've shown you the word, I'll "
print "wait for it to be completely silent, then I'll tell you. You can then "
print "say the word. I'll again wait for it to be completely silent, then I'll "
print "save what you said. If you want to pause at any time, just remain silent "
print "when I ask you to say the next word; I'll wait 5 seconds and then exit. "
print "You can pick up where you left off by running this program again."
print ""
print "Before we start, I need to figure out what silence in your environment "
print "sounds like. Make sure you don't talk, but that any ambient noise sources "
print "(such as fans) are on so I can see what they sound like. When you're ready, "
print "press enter. I'll get an idea of silence for about five seconds and then "
print "tell you. So, press enter when you're ready. But make sure not to talk!"
raw_input()

print "Waiting two seconds for the sound of you pressing enter to pass..."
time.sleep(2)
print "Starting your microphone up..."
block_size = 256
block_multiplier = 4

pa = pyaudio.PyAudio()
mic = pa.open(format=pyaudio.paInt16, channels=1, rate=44100, input=True, output=False, frames_per_buffer=block_size)
mic.start_stream()
print "Awesome, your microphone works. Let's see what silence sounds like..."

silent_blocks = []
for i in xrange(200*block_multiplier):
    silent_blocks.append(mic.read(block_size))

mic.stop_stream()
print "Hang on a second while I analyze what I just listened to..."
silence = 0
for block in silent_blocks:
    for i in xrange(0, len(block), 2):
        value = ord(block[i]) | (ord(block[i + 1]) << 8)
        if(value > 32767):
            value = value - 65536
        silence += value if value > 0 else value * -1
silence = silence / (200 * block_size)
del silent_blocks
    
print "Silence level is " + str(silence)
print ""
print "Awesome, I now know what silence sounds like in your environment. Now I "
print "need to know what it sounds like when you're talking. Press enter, then "
print "start talking about anything you want until I tell you to stop. Make sure "
print "that you talk the entire time, and don't leave any gaps while you're talking. "
print "It'll be about 5 seconds. You could read this message out loud if you want. So,"
print "press enter when you're ready."
raw_input()

print "Waiting two seconds for the sound of you pressing enter to pass..."
time.sleep(2)
print "Now let's see what you sound like when you're talking..."

noisy_blocks = []
mic.start_stream()
for i in xrange(200*block_multiplier):
    noisy_blocks.append(mic.read(block_size))

mic.stop_stream()
print "Hang on a second while I analyze what I just listened to..."
noise = 0
for block in noisy_blocks:
    for i in xrange(0, len(block), 2):
        value = ord(block[i]) | (ord(block[i + 1]) << 8)
        if(value > 32767):
            value = value - 65536
        noise += value if value > 0 else value * -1
noise = noise / (200 * block_size)
del noisy_blocks

print "Noisy level is " + str(noise)
print ""

noise_threshold = silence + ((noise - silence) / 15) 

print "Using threshold " + str(noise_threshold)
print ""
print "Awesome, I now know about what it sounds like when you're talking. Let's"
print "start recording. I'll tell you that a particular word is next. Wait a bit in "
print "silence until I tell you to say the word, then say it. If you don't like "
print "your recording of a particular word, just keep going; you can delete that "
print "word's file in your voice folder late and re-run this program to re-record it."
print "I'll keep doing this over and over again until you've recorded all of the words "
print "in the wordlist file you specified. Again, if you want to pause, just clam up "
print "for five seconds and I'll exit, and you can re-run this program when you want "
print "to resume."
print ""
print "Press enter when you're ready to start."
raw_input()
print "Waiting 3 seconds..."
time.sleep(3)

# Now we start.

silence_before = 3
silence_after = 5

word_count = len(new_words)
for index, current_word in enumerate(new_words):
    if (index % 20) == 0 and index != 0:
        print "Awesome. 20 more words down. Press enter when you're ready to keep going."
        raw_input()
        print "Waiting 3 seconds..."
        time.sleep(3)
    print "Next word is: (" + str(index + 1) + " of " + str(word_count) + ")   --->   " + current_word
    mic.start_stream()
    recent_blocks = []
    recent_noise = []
    while len(recent_blocks) < 3 + silence_before or not last_are_silent(len(recent_blocks)):
        read_and_rotate()
    # It's silent now.
    print "Say the word above."
    read_since_start = 0
    while read_since_start < 200 and not last_are_noisy(5):
        read_and_rotate()
        read_since_start += 1
    if read_since_start >= 200:
        print "You were silent for five seconds. I'll assume this means you want to"
        print "pause. Run this program again when you want to resume. Good-bye!"
        sys.exit()
    recording_buffer = recent_blocks[-(5 + silence_before):]
    while not last_are_silent(10):
        read_and_rotate()
        recording_buffer.append(recent_blocks[-1])
    print "Saving..."
    recording_buffer = recording_buffer[:-(10 - silence_after)]
    mic.stop_stream()
    file_name = os.path.join(voice_folder, current_word + ".wav")
    wave_file = wave.open(file_name, "w")
    wave_file.setnchannels(1)
    wave_file.setsampwidth(2)
    wave_file.setframerate(44100)
    wave_file.writeframes("".join(recording_buffer))
    wave_file.close()

print "Done!"





















import autobus
import special_interface

def interface_count():
    """
    An integer representing the number of interfaces currently registered to
    the Autobus server.
    """
    return len(autobus.interface_map)

def interfaces():
    """
    A list of maps representing all of the interfaces currently registered.
    This object's value is the value that is returned from list_interfaces().
    """
    return special_interface.interface.list_interfaces()

def interface_items():
    """
    A list of maps representing all of the interfaces currently registered,
    along with their functions. This is the same as the interfaces object,
    except that each map representing an interface has three additional keys,
    functions, events, and objects, whose values are the return values of the
    corresponding list_* functions for that particular interface.
    """
    interfaces = special_interface.interface.list_interfaces()
    for interface in interfaces:
        interface["functions"] = special_interface.interface.list_functions(interface["name"])
        interface["events"] = special_interface.interface.list_events(interface["name"])
        interface["objects"] = special_interface.interface.list_objects(interface["name"])
    return interfaces
        















from libautobus import AutobusConnection
import sys
import gobject
import gtk

gobject.threads_init()
    
class UserMessageInterface(object):
    def show(self, text):
        pass

def main():
    if len(sys.argv) < 2:
        print "You need to specify the name for this usermessage daemon."
        sys.exit()
    bus = AutobusConnection()
    
    





from concurrent import AtomicInteger

id_sequence = AtomicInteger(1)
import time
import socket
import random


def create_id():
    return str(time.time()) + "-" + str(id_sequence.get_and_add())


def create_service_id():
    return create_id() + "-" + socket.gethostname() + "-" + str(random.random())


def create_command(command, send_as_notice=False, **kwargs):
    result = {"_id": create_id(), "_type": 3 if send_as_notice else 1,
            "_command": command}
    result.update(kwargs)
    return result


def create_response(command_or_id, **kwargs):
    if isinstance(command_or_id, dict) and command_or_id.get("_type") == 3:
        # This is a notice, and notices aren't ever supposed to be responded to
        return None
    result = {"_id": command_or_id if isinstance(command_or_id, basestring)
              else command_or_id["_id"], "_type": 2}
    result.update(kwargs)
    return result


def create_error(command_or_id, reason):
    return create_response(command_or_id, _error={"text": reason})


def convert_to_notice(command):
    command["_type"] = 3
    return command











from autobus2 import Bus
import sys
from time import sleep
from libstatmonitor import CPUMonitor, MemoryMonitor
import socket

def main():
    global status_object
    global cpu_monitor
    global memory_monitor
    if len(sys.argv) <= 1:
        host = socket.gethostname()
    else:
        host = sys.argv[1]
    with Bus() as bus:
        service = bus.create_service({"type": "monitord", "name": host})
        status_object = service.create_object("status", None, doc=
                "Provides system status information such as current CPU "
                "usage, memory usage, etc, about this machine.")
        service.activate()
        cpu_monitor = CPUMonitor()
        cpu_monitor.refresh()
        memory_monitor = MemoryMonitor()
        sleep(0.2)
        update_status()
        try:
            while True:
                sleep(5)
                update_status()
        except KeyboardInterrupt:
            print "Interrupted, shutting down"


def update_status():
    cpu_monitor.refresh()
    memory_monitor.refresh()
    cpu_list = [cpu_monitor.get_condensed(cpu) for cpu 
            in range(cpu_monitor.get_processors())]
    memory_list = memory_monitor.get()
    status_object.set_value({"cpu_i": cpu_list, "cpu":
            cpu_monitor.get_condensed(), "memory": memory_list})





























import autobus2
import time

def main():
    with autobus2.Bus() as bus:
        service = bus.create_service({"autobus.example": "testobjectserver1"})
        object = service.create_object("test", time.time())
        service.activate()
        while True:
            try:
                time.sleep(3)
                object.set_value(time.time())
            except KeyboardInterrupt:
                return




from autobus2 import DISCOVERED, CHANGED, UNDISCOVERED
import time 

def print_service_listener(service_id, host, port, info, event):
    print time.ctime(time.time()) + ": SERVICE " + str(event).rpartition(".")[2] + ": " + service_id + " on " + host + ":" + str(port) + " -- " + str(info)



from ConfigParser import RawConfigParser
from threading import Thread, RLock
import sys
import os
import os.path
import wave
import pyaudio
import random
from SimpleXMLRPCServer import SimpleXMLRPCServer as XMLRPCServer
from SimpleXMLRPCServer import SimpleXMLRPCRequestHandler as XMLRPCRequestHandler
import time
from itertools import chain
import re
from datetime import datetime
from autobus2 import Bus
from concurrent import synchronized
from cStringIO import StringIO
import audioop

lock = RLock()

million = 1000 * 1000
billion = 1000 * 1000 * 1000
trillion = 1000 * 1000 * 1000 * 1000

print "Speakd version 9.0.0.1, starting up..."
print ""
print "    ...what, nine thousand? There's no way that can be right..."
print ""

configuration = RawConfigParser()
print "Loading configuration from speakd.conf..."
try:
    configuration.read("speakd.conf")
except:
    print "You need to create the configuration file speakd.conf before you "
    print "can use speakd. Open speakd.conf.example for an example."
    sys.exit()

voices = {}
default_voice = None
speech_queue = {}
speech_queue_lock = RLock()
is_shutting_down = False

class Sentence(object):
    def __init__(self, components, voice, priority):
        self.components = components
        self.voice = voice
        self.priority = priority

class Voice(object):
    def __init__(self, name):
        self.name = name
        self.path = configuration.get("voice:" + name, "path")
        self.scale_level = None
        if configuration.has_option("voice:" + name, "scale"):
            self.scale_level = configuration.getfloat("voice:" + name, "scale")

class VoiceChange(object):
    def __init__(self, voice):
        self.voice = voice
    def __repr__(self):
        return "<VoiceChange: " + self.voice + ">"

class ScaleLevel(object):
    def __init__(self, level, word):
        self.level = float(level)
        self.word = word
    def __repr__(self):
        return "<ScaleLevel: " + str(self.level) + " " + self.word + ">"

@synchronized(lock)
def get_next_sentence():
    if len(speech_queue) == 0:
        return None
    priority = None
    for current_priority in speech_queue:
        if priority is None or current_priority > priority:
            priority = current_priority
    speech_list = speech_queue[priority]
    result = speech_list[0]
    del speech_list[0]
    if len(speech_list) == 0:
        del speech_queue[priority]
    return result

print "Loading voices..."

for section in configuration.sections():
    if not section.startswith("voice:"):
        continue
    voice_name = section[len("voice:"):]
    voices[voice_name] = Voice(voice_name)
    if default_voice is None:
        default_voice = voice_name

if default_voice is None:
    print "You don't have any voices in your configuration file."
    sys.exit()

print "Starting the RPC server..."

def format_n(number):
    if isinstance(number, basestring):
        number = int(number)
    if number <= 20:
        return str(number)
    if number < 100:
        if (number % 10) == 0:
            return str(number)
        else:
            return str(number / 10 * 10) + " " + str(number % 10)
    if number < 1000:
        if (number % 100) == 0:
            return str(number / 100) + " hundred"
        else:
            return str(number / 100) + " hundred " + format_n(number % 100)
    if number < million:
        if (number % 1000) == 0:
            return format_n(number / 1000) + " thousand"
        else:
            return format_n(number / 1000) + " thousand " + format_n(number % 1000)
    if number < billion:
        if (number % million) == 0:
            return format_n(number / million) + " million"
        else:
            return format_n(number / million) + " million " + format_n(number % million)
    if number < trillion:
        if (number % billion) == 0:
            return format_n(number / billion) + " billion"
        else:
            return format_n(number / billion) + " billion " + format_n(number % billion)
    raise Exception("No such number")

def format_nd(number):
    result = format_n(number)
    suffix = result.split(" ")[-1]
    if suffix == "1":
        return result + "st"
    elif suffix == "2":
        return result + "nd"
    elif suffix == "3":
        return result + "rd"
    return result + "th"

def format_p(number):
    return int(number)

def format_t(time_string):
    hour, minute, period = re.match("([0-9]?[0-9]):([0-9]?[0-9])"
            "((?:[AaPp]\\.?[Mm]\\.?)?)", time_string).groups()
    hour = int(hour)
    minute = int(minute)
    result = format_n(hour).split(" ")
    if minute == 0:
        result.append("o'clock")
    elif minute < 10:
        result += ["oh"] + format_n(minute).split(" ")
    else:
        result += format_n(minute).split(" ")
    if period != "":
        period = period.lower()
        if period[0] == "a":
            result.append("am")
        else:
            result.append("pm")
    return result

def format_pt(data):
    singular, suffix, count = data.split(":")
    count = int(count)
    if count == 1:
        return singular
    else:
        return singular + suffix

def format_v(voice_string):
    if voice_string == "":
        voice_string = random.choice(voices.keys())
    if voice_string not in voices:
        raise Exception("No such voice: " + voice_string)
    return VoiceChange(voice_string)

def format_sl(scale_string):
    return ScaleLevel(*scale_string.split(":", 1))

def scale_sound(block, scale_level):
    return audioop.mul(block, 2, scale_level)

class RPC(object):
    """
    Allows communicating with the speak server. This allows pre-recorded
    phrases and synthesized speech to be spoken.
    """
    @synchronized(lock)
    def say(self, components, priority=0, voice=None):
        """
        NOTE: Usually you'll want to use say_text instead of this function.
        That said, here's the documentation...
        
        Adds the specified sentence to the queue. It will be spoken as soon as
        there are no more sentences with a priority higher than this
        sentence's priority and all previously-added sentences with the same
        priority have been spoken. The first parameter, components, is a list
        of strings and ints. Each string represents a single sound to say,
        while each int represents a number of milliseconds to pause before
        continuing in the sentence. Pause gaps can be no longer than 5 seconds.
        The second parameter is an int representing the priority of the
        sentence. Higher-priority sentences will be spoken first. The third,
        and optional, parameter, specifies the name of the voice that should
        be used to say the sentence. If this is null, the sentence will be
        spoken with whatever the default voice is at the time that the speak
        server actually starts to speak the sentence.
        """
        assert isinstance(priority, int)
        if voice is not None and voice not in voices:
            raise Exception("There is no voice with that name. Voices must be "
                            "configured in speakd.conf.")
        if priority not in speech_queue:
            speech_queue[priority] = []
        speech_queue[priority].append((priority, Sentence(components, voice, priority)))
        print "After adding request, speech queue is " + str(speech_queue)
#        speech_queue_object.set(speech_queue)
    
    @synchronized(lock)
    def say_text(self, text, priority=0, voice=None):
        """
        Same as say(text, priority=0, voice=None), but text should be a string,
        and it is parsed into tokens around space characters. If a token starts
        with a : character, then it should be of the format ":type:data", where
        type is a type specifier and data is additional data for that type.
        Valid type specifiers are:
        
        n: The data is a number. This token will be replaced with the words
        needed to say that particular number. Negative numbers are not
        currently allowed. For example, ":n:123" would be replaced with
        "1 hundred 20 3".
        
        nd: The data is a number. This token will be replaced with the words
        needed to state the position of an item by that number. For example,
        ":nd:1" would be replaced with "1st", ":nd:2" would be replaced with
        "2nd", ":nd:123" would be replaced with "1 hundred 20 3rd", and so on.
        
        p: The data is a number. This token causes a pause of the specified
        number of milliseconds to occur.
        
        t: The data is a time, in the format 12:34pm or similar. The period
        specifier can be a.m., p.m., am, or pm, and both uppercase and
        lowercase versions are allowed. 12:34pm would be expanded to
        "12 30 4 pm", 5:03pm (or 05:03pm) would be expanded to "5 oh 3 pm",
        and 7:00pm would be expanded to "7 o'clock pm".
        
        pt: The data is a string of the form "singular:suffix:count". singular
        is the singular (in terms of the English language) form of a word. 
        suffix is the suffix that should be appended to the singular form of
        the word to turn it into its plural (this is usually "s" for most
        words). count is a number. If count is not numerically equal to the
        number 1, this token expands into its plural form (I.E. the singular
        with the suffix appended). Otherwise, this token expands into the
        singular form of the word. 
        """
        print "Received request to say text " + str(text) 
        tokens = text.split(" ")
        new_tokens = []
        for token in tokens:
            if not token:
                pass
            elif token[0] == ":":
                specifier, data = token[1:].split(":", 1)
                format_result = globals()["format_" + specifier](data)
                if isinstance(format_result, basestring):
                    new_tokens += format_result.split(" ")
                elif isinstance(format_result, list):
                    new_tokens += format_result
                else:
                    new_tokens.append(format_result)
            else:
                new_tokens.append(token)
        self.say(new_tokens, priority, voice)
    
    @synchronized(lock)
    def get_queue_size(self):
        """
        Returns the number of sentences currently in the queue. This does not
        include the sentence that is currently being spoken if there is one.
        """
        return len(list(chain(*speech_queue.values())))
    
    @synchronized(lock)
    def get_voice_names(self):
        """
        Returns a list of the names of all of the voices available to this
        speak server.
        """
        return voices.keys()
    
    @synchronized(lock)
    def get_default_voice(self):
        """
        Returns the name of the default voice. The default voice is the voice
        used to speak sentences when no voice is specified for the sentence.
        When the speak server starts up, the default voice will be the first
        voice in the configuration file.
        """
        return default_voice
    
    @synchronized(lock)
    def set_default_voice(self, voice_name):
        """
        Sets the default voice. The default voice is the voice used to speak
        sentences when no voice is specified for the sentence.
        """
        global default_voice
        if voice_name not in voices:
            raise Exception("There is no voice with that name. Voices must be "
                            "configured in speakd.conf.")
        default_voice = voice_name
    
    @synchronized(lock)
    def get_pid(self):
        """
        Returns the pid of the speak server.
        """
        return os.getpid()


def sanitize_file(name):
    return name.replace(".", "").replace("\\", "").replace("/", "")


def main():
    global audio_device
    global autio_stream
    global bus
    global speech_queue_object
    global speech_queue_size_object
    audio_device = pyaudio.PyAudio()
    audio_stream = audio_device.open(format=audio_device.get_format_from_width(2),
                                     channels=1, rate=44100, output=True,
                                     frames_per_buffer=1024)
    audio_stream.stop_stream()
    
    with Bus() as bus:
        service = bus.create_service({"type": "speak"}, active=False, from_py_object=RPC())
        speech_queue_object = service.create_object("speech_queue",
        {}, "This doesn't work yet.")
        speech_queue_size_object = service.create_object(
                "speech_queue_size", 0, "The number of items currently in the "
                "speech queue. This does not include the item currently being "
                "spoken, if any.")
        service.activate()
        
        print ""
        print "Speakd has successfully started up. PID is " + str(os.getpid())
        try:
            while not is_shutting_down:
            #   print "Main loop"
                sentence_tuple = get_next_sentence()
                if sentence_tuple is None:
            #       print "Nothing to say"
                    time.sleep(0.3)
                    continue
                priority, sentence = sentence_tuple
                print "Something to say: " + str(sentence.components)
                if sentence.voice is not None:
                    voice = voices[sentence.voice]
                else:
                    voice = voices[default_voice]
                print "Saying with voice " + voice.name
                for item in sentence.components:
                    scale_level = 1.0
                    if voice.scale_level is not None:
                        scale_level = scale_level * voice.scale_level
                    if isinstance(item, ScaleLevel):
                        scale_level = scale_level * item.level
                        item = item.word
                    if isinstance(item, basestring):
                        audio_stream.start_stream()
                        file = sanitize_file(item)
                        print "Saying word " + file
                        absolute_file = os.path.join(voice.path, file + ".wav")
                        try:
                            wave_file = wave.open(absolute_file, "r")
                        except:
                            print "File " + absolute_file + " does not exist for voice " + voice.name + ". It will be silently ignored."
                            continue
                        data = wave_file.readframes(1024)
                        while data != "":
                            audio_stream.write(scale_sound(data, scale_level))
                            data = wave_file.readframes(1024)
                        audio_stream.stop_stream()
                    elif isinstance(item, int):
                        print "Pausing for " + str(item) + " milliseconds"
                        time.sleep((item * 1.0) / 1000.0)
                    elif isinstance(item, VoiceChange):
                        print "Switching voice to " + item.voice
                        voice = voices[item.voice]
                print "Waiting..."
                time.sleep(1)
                print "Closing audio stream"
                print "Done."
        except KeyboardInterrupt:
            print "Interrupted, shutting down"






















import code, readline
import librtk
import signal

def main():
    print "Listening on port 6785 for a RTK connection."
    c=librtk.listen(6785)
    print "Connection received, and declared as `c`"
    
    try:
        code.InteractiveConsole({"c": c}).interact()
    finally:
        c.close()

if __name__ == "__main__":
    main()




import os
import os.path

def native_path(path):
    return path.replace("/", os.sep)


def internal_path(path):
    return path.replace(os.sep, "/")


class File(object):
    def __init__(self, path):
        while path.endswith("/"):
            path = path[:-1]
        self.path = path
        self.native_path = native_path(path)
    
    @property
    def is_folder(self):
        return os.path.isdir(self.native_path)
    
    @property
    def is_file(self):
        return os.path.isfile(self.native_path)
    
    @property
    def exists(self):
        return os.path.exists(self.path)
        
    def check_folder(self):
        if not self.is_folder:
            raise Exception('"%s" does not exist or is not a directory' % self.path)
    
    def check_file(self):
        if not self.is_file:
            raise Exception('"%s" does not exist or is not a file' % self.path)
    
    def list(self):
        results = os.listdir(self.native_path)
        return [File(self.path + "/" + p) for p in results]
    
    def open(self, *args, **kwargs):
        return open(self.native_path, *args, **kwargs)
    
    def child(self, path):
        return File(self.path + "/" + path)
    
    def parent(self):
        return File(internal_path(os.path.dirname(self.native_path)))
    
    @property
    def name(self):
        return os.path.split(self.path)[1]
    
    def recurse(self, filter=None):
        """
        A generator that recursively yields all child File objects of this file.
        Files and directories (and the files and directories contained within
        them, and so on) are all included; files for which the specified filter
        function return False (or None or some other Python false-like value)
        are not included.
        
        Note that directories for which the filter returns False are still
        recursed into. I might change this later to allow specifying whether
        the directory should not be recursed into when it's filtered out.
        """
        children = self.list()
        for child in children:
            if filter is None or filter(child):
                yield child
            if child.is_folder:
                for c in child.recurse(filter):
                    yield c
    
    def __str__(self):
        return "afn.fileutils.File(%r)" % self.path
    
    __repr__ = __str__

    




































from abc import ABCMeta as ABC, abstractmethod as abstract
import autobus2
import inspect

class AutoClose(object):
    """
    A class intended to be used as a superclass by classes with a close()
    function. AutoClose turns classes that subclass it into context managers
    whose __exit__ functions cause self.close() to be run. The __enter__ and
    __exit__ provided by AutoClose are reentrant.
    
    AutoClose's __enter__ returns self.
    """
    def __enter__(self):
        try:
            self._autoclose_enters += 1
        except AttributeError:
            self._autoclose_enters = 1
        return self
    
    def __exit__(self, *args):
        self._autoclose_enters -= 1
        if self._autoclose_enters == 0:
            self.close()


class ServiceConnector(object):
    """
    This class isn't yet complete. I'm hoping to use it to merge the stuff in
    Connection, SingleServiceProxy, and MultipleServiceProxy into this class at
    some point in the future.
    """
    __metaclass__ = ABC
    
    @abstract
    def _call_function(self, args, callback, timeout, safe):
        pass
    
    def __getitem__(self, name):
        return self.functions[name]
    
    def _sc_get_function(self, name):
        return ServiceConnectorFunction(self, name)
    
    @property
    def functions(self):
        # TODO: cache this
        return ServiceConnectorAccessor(self._sc_get_function)


class ServiceConnectorAccessor(object):
    def __init__(self, accessor):
        self.accessor = accessor
    
    def __getattr__(self, name):
        return self.accessor(name)
    
    def __getitem__(self, name):
        return self.accessor(name)


class ServiceConnectorFunction(object):
    def __init__(self, connector, name):
        self.connector = connector
        self.name = name
    
    def __call__(self, *args, **kwargs):
        callback = kwargs.get("callback", autobus2.SYNC)
        timeout = kwargs.get("timeout", 30)
        safe = kwargs.get("safe", False)
        return self.connector._call_function(args, callback, timeout, safe)


def get_function_doc(object, function_name, remove_self=True):
    function = getattr(object, function_name)
    is_method_before = inspect.ismethod(function)
    while hasattr(function, "wrapped"):
        function = function.wrapped
    if is_method_before or inspect.ismethod(function):
        argspec = inspect.getargspec(function)
        # Remove the leading "self" argument
        if remove_self:
            argspec = (argspec[0][1:],) + argspec[1:]
        args = inspect.formatargspec(*argspec)
    elif inspect.isfunction(function):
        args = inspect.formatargspec(*inspect.getargspec(function))
    else:
        args = "(...)"
    doc = inspect.getdoc(function)
    if doc is None:
        doc = ""
    return function_name + args + "\n\n" + doc








import librtk
import librtk.protocols
import librtkinter
from ConfigParser import RawConfigParser
import sys
import utils
import time

# A list of lists of components for each key. The sublists are [row, col, color]
# specifying where in the key's table the text field for that component should
# be placed and the background color for the text field. The first
# element in the overall list represents the location of
# the component for lowercase chars, the second uppercase, the third altgr,
# and the fourth altgr uppercase (I.E. altgr+shift). Once I figure out how to
# add additional components with XKB groups, I'll likely add more entries.
KEY_COMPONENTS = [[3, 0, "#ffffff"], [2, 0, "#ccccff"],
                  [3, 1, "#ffcccc"], [2, 1, "#ffccff"],
                  [1, 0, "#ffffc5"], [0, 0, "#ccffcc"],
                  [1, 1, "#ffeacc"], [0, 1, "#eeeeee"]]
# The names of each component. These contain representatiosn of what key
# combinations cause that component to be used.
KEY_COMPONENT_NAMES = ["none", "shift",
                       "altgr", "altgr+shift",
                       "ctrl", "ctrl+shift",
                       "ctrl+altgr", "ctrl+altgr+shift"]

# A list of rows. Each entry in this list has four items: the row letter in
# the XKB file, the row number in the RTK table, the offset from the left of
# the RTK table at which this row should start, and the number of keys present
# in this row.
KEY_ROWS = [["B", 3, 1, 10], ["C", 2, 1, 11], ["D", 1, 1, 13], ["E", 0, 0, 13]]

def main():
    global config, config_file, output_file, output_name, output_desc
    global connection
    if len(sys.argv) <= 4:
        print "Syntax: keyeditor <config-filename> <output-filename> <name> <desc>"
        print "The specified config file will be created if it doesn't exist."
        print "The output file will be written when the generate button is"
        print "clicked. It can then be copied into /usr/share/X11/XKB/symbols."
        return
    config_file = sys.argv[1]
    output_file = sys.argv[2]
    output_name = sys.argv[3]
    output_desc = sys.argv[4]
    config = RawConfigParser()
    config.read(config_file)
    client_protocol = librtk.protocols.LinkedProtocol()
    server_protocol = librtk.protocols.LinkedProtocol(client_protocol)
    connection = librtk.Connection(server_protocol, start_app)
    connection.start()
    client_connection, master = librtkinter.start_connection(client_protocol)
    print "Started up!"
    try:
        master.mainloop()
    except KeyboardInterrupt:
        print "Interrupted, shutting down"
        client_connection.close()
        connection.close()
    print "Terminated."

def start_app(*args): # We don't need the connection since it's stored in the
    # global field connection in main()
    global save_button, generate_button
    window = connection.Window(title="KeyEditor")
    def on_close():
        save()
        connection.close()
    window.close_request.listen(on_close)
    box = connection.VBox(window)
    key_table = connection.Table(box)
    connection.Label(box, text=" ") # Spacer
    buttons = connection.HBox(box)
    save_button = connection.Button(buttons, text="Save", clicked=save)
    generate_button = connection.Button(buttons, text="Generate", clicked=generate)
    connection.Label(buttons, text="    ") # Spacer
    create_legend(buttons)
    for key_row in KEY_ROWS:
        row_name, row, col_offset, keys = key_row
        for i in range(keys):
            col = i + col_offset
            key_widget = create_key(key_table, row_name + "-" + str(i))
            key_widget.row = row
            key_widget.col = col

def create_legend(parent):
    table = connection.Table(parent)
    for (row, col, color), name in zip(KEY_COMPONENTS, KEY_COMPONENT_NAMES):
        connection.Label(table, text=name, row=row, col=col,
                background=color, pin="we")

def create_key(parent, keyname):
    panel = connection.BorderPanel(parent)
    table = connection.Table(panel)
    for index, (row, col, color) in enumerate(KEY_COMPONENTS):
        create_key_component(keyname, table, index, row, col, color)
    return panel

def create_key_component(keyname, table, index, row, col, color):
    field = connection.TextBox(table, row=row, col=col, width=2,
            background=color, font_size=14)
    if config.has_option(keyname, str(index)):
        value = config.get(keyname, str(index))
        if value:
            field.set_text(unichr(int(value)))
    def changed():
        if len(field.text) > 1:
            print "Field length too large"
            field.set_text(field.text[0])
            return
        print ("Setting " + keyname + " component " + str(index) + 
                " to " + field.text + "")
        if not config.has_section(keyname):
            config.add_section(keyname)
        config.set(keyname, str(index), (field.text and 
                str(ord(field.text))) or "")
    field.text_changed.listen(changed)

def save():
    print "Saving..."
    with open(config_file, "w") as file:
        config.write(file)
    print "Saved!"
    save_button.text = "Saved!"
    def reset_save_button():
        save_button.text = "Save"
    utils.at(2, reset_save_button)

def generate():
    print "Generating..."
    with open(output_file, "w") as file:
        file.write('xkb_symbols "' + output_name + '"\n{\n')
        file.write('    name[Group1] = "' + output_desc + ' - 1";\n')
        file.write('    name[Group2] = "' + output_desc + ' - 2";\n')
        for row_name, _, _, keys in KEY_ROWS:
            for index in range(keys):
                keyname = row_name + "-" + str(index)
                output_key = get_output_key(row_name, index)
                file.write("    key " + output_key + " { [ ")
                file.write(generate_key_range(keyname, 0, 4))
                file.write(" ], [], [ ")
                file.write(generate_key_range(keyname, 4, 8))
                file.write(" ] };\n")
        file.write("};\n")
    print "Generated!"
    generate_button.text = "Generated!"
    def reset_generate_button():
        generate_button.text = "Generate"
    utils.at(2, reset_generate_button)

def generate_key_value(keyname, index):
    """
    Generates the XKB value of the specified component of the specified key.
    For example, on a typical layout, generate_key_value("C-0", 0) would
    return "U0061" (the unicode value for "a"). This function will return
    "NoSymbol" if there's no such key in the configuration or if the specified
    key does not have a component with the specified index.
    """
    if (config.has_section(keyname) and 
            config.has_option(keyname, str(index)) and
            config.get(keyname, str(index))):
        return ("U" + hex(config.getint(keyname, str(index)))[2:]
                .rjust(4, "0").upper())
    else:
        return "NoSymbol"

def generate_key_range(keyname, start, end):
    """
    Generates a string containing the values returned by generate_key_value
    for every component with an index between start, inclusive, and end,
    exclusive, for the specified key, with each value separated by a comma.
    """
    keycodes = []
    for i in range(start, end):
        keycodes.append(generate_key_value(keyname, i))
    return ", ".join([k.rjust(12) for k in keycodes])

def get_output_key(row_name, index):
    index += 1
    if row_name == "E" and index == 1:
        return "<TLDE>"
    if row_name == "D" and index == 13:
        return "<BKSL>"
    if row_name == "E":
        index -= 1
    return "<A" + row_name + str(index).rjust(2, "0") + ">"






















"""
This module has been moved to afn.utils. You may still use this module (it now
contains only one line, namely "from afn.utils import *", so everything's still
usable from here), but you should probably use afn.utils in any future projects
you're writing that might otherwise make use of this module.
"""

from afn.utils import * #@UnusedWildImport












from threading import Thread
from Queue import Queue, Empty
from traceback import print_exc

class TkThread(object):
    """
    This class starts a thread that runs a Tkinter mainloop(). It allows UI
    tasks to be submitted to it, which it will execute as fast as it can and
    in order. The return value from those tasks can be captured if needed. This
    makes it a bit easier for multithreaded applications to use Tkinter.
    """
    def __init__(self, init_function):
        """
        Creates a new TkThread. init_function is a function that, when called,
        should create the root Tk window and return it.
        
        After you create a new TkThread, you need to call its start function
        to actually get things going. The thread will then call the init
        function
        """
        self.init_function = init_function
        self.queue = Queue()
    
    def run(self):
        toplevel = self.init_function()
        self.toplevel = toplevel
        toplevel.after(100, self._process_events)
        toplevel.mainloop()
    
    def _process_events(self):
        try:
            while True:
                event = self.queue.get_nowait()
                try:
                    event()
                except:
                    print_exc()
        except Empty:
            pass
        finally:
            self.toplevel.after(100, self._process_events)
    
    def call(self, function, *args):
        """
        Schedules the specified function to be called on the Tkinter thread.
        """
        self.queue.put(lambda: function(*args))
    
    def call_sync(self, function, *args):
        """
        Schedules the specified function to be called on the Tkinter thread,
        then waits until it gets called and returns its return value.
        """
        wait_queue = Queue()
        def event():
            try:
                value = (False, function(*args))
            except Exception, e:
                value = (True, e)
            wait_queue.put(value)
        self.call(event)
        excepted, value = wait_queue.get()
        if excepted:
            raise value
        else:
            return value










from autobus2 import Bus

def main():
    with Bus() as bus:
        with bus.get_service_proxy({"autobus.example": "hello_server"}) as service:
            service.wait_for_bind(timeout=2)
            print service["hi"]("great big round world")




from collections import namedtuple as _namedtuple

WidgetSchema = _namedtuple("WidgetSchema", "name doc type widget_properties "
        " layout_properties state_properties calls events")
WidgetPropertySchema = _namedtuple("WidgetPropertySchema", "name doc writable default")
LayoutPropertySchema = _namedtuple("LayoutPropertySchema", "name doc writable default")
StatePropertySchema = _namedtuple("StatePropertySchema", "name doc default")
CallSchema = _namedtuple("CallSchema", "name doc")
EventSchema = _namedtuple("EventSchema", "name doc")




from socket import AF_INET, SOCK_DGRAM, socket as Socket
from dns.message import from_wire

def main():
    s = Socket(AF_INET, SOCK_DGRAM)
    s.bind(("", 53))
    while True:
        message = s.recv(1500)
        print repr(from_wire(message))
    



from __future__ import with_statement
import librtk
import librtk.protocols

"""
This package contains a number of RTK example applications. All of them are
provided as modules that contain one important function, start, which should
be passed as the on_connect function to an RTK connection.

This module's serve_example function can be used to automatically start up a
server serving a particular named example.
"""


def serve_example(port, name, localhost_only=True):
    """
    Creates a server that can serve the specified example. This function itself
    does not start the server; you need to call the return value's start
    function to start the server. The server will be a ThreadedServer, so if
    you'd like to run the accept loop on the current thread you can call the
    run function on the return value instead of the start function.
    
    If localhost_only is True (the default), the example will be served only
    on the loopback interface. If it's false, the example will be served on
    all interfaces.
    """
    __import__("librtk.examples." + name)
    import sys
    return librtk.protocols.ThreadedServer("127.0.0.1" if localhost_only else "", 
            port, sys.modules["librtk.examples." + name].start)

def main():
    """
    Calls serve_example, taking the port and name from the first two
    command-line arguments. This is intended for use with afn-python/run.
    
    The third command-line argument, which is optional and should be true or
    false (the default), specifies whether or not the example should be
    accessible outside of the local host.
    """
    import sys
    if len(sys.argv) <= 2:
        print "You need to specify a port and an example name."
        return
    server = serve_example(int(sys.argv[1]), sys.argv[2],
            localhost_only=len(sys.argv) > 3 and sys.argv[3] == "true")
    try:
        server.run()
    except KeyboardInterrupt:
        print "Interrupted, shutting down"
    finally:
        server.shutdown()
        sys.exit()














from __future__ import with_statement

from threading import Thread, RLock
import threading
from functools import partial, update_wrapper
import inspect
import sys
import traceback
import afn.utils

def as_new_thread(function): 
    """
    A decorator that causes the decorated function to be invoked in a new
    thread. The function returned will invoke the specified function on a
    newly-created thread. The target function's return value will be discarded;
    the resulting function will always return None.
    """
    def wrapper(*args, **kwargs):
        Thread(name="afn.utils.concurrent.as_new_thread", target=function, args=args, kwargs=kwargs).start()
    update_wrapper(wrapper, function)
    wrapper.wrapped = function
    return wrapper

def synchronized(lock=None, function=None):
    """
    Returns a function wrapping the specified function that locks on the
    specified lock before calling the function and releases it after the 
    function has been invoked. Omitting the second argument will return
    functools.partial(synchronized, lock), and specifying a callable lock will
    instead return synchronized(None, lock), which effectively allows this
    function to be used as a decorator thus:
    
    @synchronized(some_lock)
    @synchronized()
    @synchronized
    
    All three uses are valid, with the latter two creating a new reentrant
    lock for this function.
    
    Technically, the specified lock can be any object that can be passed as the
    context manager of a with statement. This behavior will be preserved
    throughout potential future revisions of this function/decorator.
    """
    if callable(lock): # in case it's used as @synchronized
        return synchronized(None, lock)
    if function is None:
        return partial(synchronized, lock)
    if lock is None:
        lock = RLock()
    def wrapper(*args, **kwargs):
        with lock:
            return function(*args, **kwargs)
    update_wrapper(wrapper, function)
    wrapper.wrapped = function
    wrapper.__doc__ = format_argspec(function) + "\n\n" + getdoc(function)
    return wrapper

def synchronized_on(attribute=None):
    """
    Similar to the synchronized function, but must be used as a decorator thus:
    
    @synchronized_on("some_attribute")
    def some_method(self, ...):
        ...
    
    Namely, the method being decorated must be just that: a method inside of a
    class. This will cause the method to be synchronized on the lock (or any
    other sort of context manager) located at self.some_attribute, which allows
    for methods to be synchronized on per-instance locks.
    
    some_attribute can contain dots; these will be parsed as per normal Python
    semantics, so, for example, @synchronized_on("x.y") synchronizes on
    self.x.y.
    """
    def decorator(function):
        def wrapper(self, *args, **kwargs):
            with afn.utils.get_path(self, attribute):
                return function(self, *args, **kwargs)
        update_wrapper(wrapper, function)
        wrapper.wrapped = function
        wrapper.__doc__ = format_argspec(function) + "\n\n" + getdoc(function)
        return wrapper
    return decorator

def format_argspec(function):
    argspec = inspect.getargspec(function)
    args = inspect.formatargspec(*argspec)
    return function.__name__ + args

def getdoc(function):
    doc = inspect.getdoc(function)
    if doc is None:
        doc = ""
    return doc

class AtomicInteger(object):
    """
    An integer or long (it switches between the two as needed) that provides
    a guarantee of atomicity across all operations it supports, and in
    particular, it guarantees that inline operations (such as += and *=) will
    be atomic. For example:
    
    number = AtomicInteger()
    number.set(1)
    number += 1
    number == 2
    """

    def __init__ (self, value=0):
        self.integer = value
        self.lock = RLock()
        return
    
    def get(self):
        with self.lock:
            return self.integer
    
    def get_and_add(self, value_to_add=1):
        with self.lock:
            value = self.integer
            self.integer += value_to_add
            return value
    
    def set(self, integer):
        with self.lock:
            self.integer = integer
    
    def __cmp__(self, other):
        with self.lock:
            return self.integer.__cmp__(other)
    
    def __hash__(self):
        with self.lock:
            return self.integer.__hash__()
    
    def __nonzero__(self):
        with self.lock:
            return self.integer.__nonzero__()
    
    def __repr__(self):
        return str(self.integer)
    
    def __int__(self):
        return int(self.integer)
    
    def __long__(self):
        return long(self.integer)
    
    def __iadd__(self, other):
        with self.lock:
            self.integer = self.integer.__add__(other)
            return self
    
    def __isub__(self, other):
        with self.lock:
            self.integer = self.integer.__sub__(other)
            return self
    
    def __imul__(self, other):
        with self.lock:
            self.integer = self.integer.__mul__(other)
            return self
    
    def __idiv__(self, other):
        with self.lock:
            self.integer = self.integer.__div__(other)
            return self
    
    def __itruediv__(self, other):
        with self.lock:
            self.integer = self.integer.__truediv__(other)
            return self
    
    def __ifloordiv__(self, other):
        with self.lock:
            self.integer = self.integer.__floordiv__(other)
            return self
    
    def __imod__(self, other):
        with self.lock:
            self.integer = self.integer.__mod__(other)
            return self
    
    def __ipow__(self, other):
        with self.lock:
            self.integer = self.integer.__pow__(other)
            return self
    
    def __ilshift__(self, other):
        with self.lock:
            self.integer = self.integer.__lshift__(other)
            return self
    
    def __irshift__(self, other):
        with self.lock:
            self.integer = self.integer.__rshift__(other)
            return self
    
    def __iand__(self, other):
        with self.lock:
            self.integer = self.integer.__and__(other)
            return self
    
    def __ixor__(self, other):
        with self.lock:
            self.integer = self.integer.__xor__(other)
            return self
    
    def __ior__(self, other):
        with self.lock:
            self.integer = self.integer.__or__(other)
            return self


def dump_thread_traces():
    """
    Prints a traceback to stdout for each live thread in the Python interpreter.
    This is useful for when infinite loops or other strange behavior are
    occurring and you're not sure where; you could write some code to call this
    function when such a situation is detected or when the user sends a
    particular signal to the process.
    """
    threads = threading.enumerate()
    threads_by_id = dict((t.ident, t) for t in threads)
    stack_traces = sys._current_frames()
    print "*** STACK TRACES OF ALL LIVE THREADS ***"
    print ""
    for thread_id, frame in stack_traces.items():
        print "Stack trace for " + str(threads_by_id.get(thread_id, thread_id)) + ":"
        lines = traceback.format_list(traceback.extract_stack(frame))
        for line in lines:
            print line,
        print ""
    print "*** END STACK TRACES ***"





















from afn.dnsd.server import Server
from afn.dnsd.data import Answer, A, CNAME

def resolver(source, question, name):
    print (source, question, name)
    print "Question: ", source, question, name
    if question == "A":
        return [
            Answer(name, "CNAME", 1, [
                CNAME("www.redirected.example.com")
            ]),
            Answer("www.redirected.example.com", "A", 1, [
                A("1.2.3.4"),
                A("5.6.7.8"),
                A("9.10.11.12")
            ])
        ]

def main():
    server = Server(port=5053)
    server.resolvers.append(resolver)
    server.run()
